"""
Nested model relationships for Aquarium models.
"""

import json

import inflection

from pydent.base import ModelBase
from pydent.marshaller import fields
from pydent.marshaller.exceptions import ModelValidationError


class FieldValidationError(ModelValidationError):
    pass


class Raw(fields.Field):
    """
    Field that performs no serialization/deserialization.
    """

    def __init__(self, many=False, data_key=None, allow_none=True, default=None):
        super().__init__(
            many=many, data_key=data_key, allow_none=allow_none, default=default
        )


class JSON(Raw):
    """
    Automatically serializes/deserializes JSON objects.
    """

    def _deserialize(self, owner, data):
        if isinstance(data, dict):
            return data
        return json.loads(data)

    def _serialize(self, owner, data):
        return json.dumps(data)


class Function(fields.Callback):
    """
    Calls a specified function upon attribute access. Similar to the @property
    decorator in python, but will search and find an instance method using
    the method name.
    """

    def __init__(
        self,
        callback,
        callback_args=None,
        callback_kwargs=None,
        cache=False,
        data_key=None,
        many=None,
        allow_none=True,
        always_dump=True,
    ):
        super().__init__(
            callback,
            callback_args,
            callback_kwargs,
            cache,
            data_key,
            many,
            allow_none,
            always_dump,
        )


class BaseRelationshipAccessor(fields.RelationshipAccessor):
    """
    Python descriptor that is returned by a field during attribute access.
    """

    HOLDER = None


class BaseRelationship(fields.Relationship):
    """
    Base class for relationships. By default, if the value is None, attempt a callback. If that fails,
    fallback to None. If successful, deserialize data to the nested model.
    """

    QUERY_TYPE = None
    ACCESSOR = BaseRelationshipAccessor

    def __init__(
        self,
        nested,
        callback,
        callback_args=None,
        callback_kwargs=None,
        many=None,
        allow_none=True,
    ):
        super().__init__(
            nested,
            callback,
            callback_args,
            callback_kwargs,
            cache=True,
            data_key=None,
            many=many,
            allow_none=allow_none,
        )

    def fullfill(self, owner, cache=None, extra_args=None, extra_kwargs=None):
        try:
            return super().fullfill(
                owner, cache, extra_args=extra_args, extra_kwargs=extra_kwargs
            )
        except fields.RunTimeCallbackAttributeError as e:
            return BaseRelationshipAccessor.HOLDER

    def build_query(self, models):
        """Bundles all of the callback args for the models into a single query"""
        args = {}
        for s in models:
            callback_args = self.get_callback_args(s)[1:]
            if self.QUERY_TYPE == "by_id":
                args.setdefault(self.attr, [])
                for x in callback_args:
                    if x is not None and x not in args[self.attr]:
                        args[self.attr].append(x)
            else:
                for cba in callback_args:
                    for k in cba:
                        args.setdefault(k, [])
                        arg_arr = args[k]
                        val = cba[k]
                        if val is not None:
                            if isinstance(val, list):
                                for v in val:
                                    if v not in arg_arr:
                                        arg_arr.append(v)
                            elif val not in arg_arr:
                                arg_arr.append(val)
        return args


class One(BaseRelationship):
    """
    Defines a single relationship with another model.
    Subclass of :class:`pydent.marshaller.Relation`.
    """

    QUERY_TYPE = "by_id"

    def __init__(
        self,
        nested,
        *args,
        callback=None,
        callback_args=None,
        callback_kwargs=None,
        **kwargs
    ):
        """
        One initializer. Uses "find" callback by default.

        :param nested: target model
        :type nested: basestring
        :param args: other args for fields.Nested relationship
        :type args: ...
        :param attr: attribute to use to find model
        :type attr: basestring
        :param kwargs: other kwargs for fields.Nested relationship
        :type kwargs: ...
        """
        if callback is None:
            callback = ModelBase.find_callback.__name__
        super().__init__(
            nested,
            callback,
            callback_args=callback_args,
            callback_kwargs=callback_kwargs,
            many=False,
        )


class Many(BaseRelationship):
    """
    Defines a many relationship with another model.
    Subclass of :class:`pydent.marshaller.Relation`.
    """

    QUERY_TYPE = "query"

    def __init__(
        self,
        nested,
        *args,
        callback=None,
        callback_args=None,
        callback_kwargs=None,
        **kwargs
    ):
        """
        Many initializer. Uses "where" callback by default.

        :param nested: target model
        :type nested: basestring
        :param args: other args for fields.Nested relationship
        :type args: ...
        :param attr: attribute to use to find model
        :type attr: basestring
        :param kwargs: other kwargs for fields.Nested relationship
        :type kwargs: ...
        """
        if callback is None:
            callback = ModelBase.where_callback.__name__
        super().__init__(
            nested,
            *args,
            many=True,
            callback=callback,
            callback_args=callback_args,
            callback_kwargs=callback_kwargs,
            **kwargs
        )


class HasMixin:
    """
    Mixin for adding the 'set_ref' method. 'set_ref' builds a 'ref' and 'attr'
    attributes
    """

    def set_ref(self, nested=None, ref=None, attr=None):
        """
        Sets the 'ref' and 'attr' attributes. These attributes are used to
        defined parameters for
        :class:`pydent.marshaller.Relation` classes.

        For example:

        .. code-block:: python

            relation # HasOne, HasMany, or HasManyGeneric, etc.
            relation.set_ref(ref="parent_id")
            relation.ref    # "parent_id"
            relation.attr   # "id"

            relation.set_ref(model="SampleType")
            relation.ref   # "sample_type_id"
            relation.attr  # "id"

            relation.set_ref(attr="name", model="OperationType")
            relation.ref   # "operation_type_name
            relation.attr  # "name"
        """
        self.ref = ref
        self.attr = attr

        if not self.attr:
            self.attr = "id"
        if ref:
            self.ref = ref
        else:
            if not self.attr:
                raise FieldValidationError(
                    "'attr' is None. Relationship '{}' needs an 'attr' and 'model' parameters".format(
                        self.__class__
                    )
                )
            if not nested:
                raise FieldValidationError(
                    "'model' is None. Relationship '{}' needs an 'attr' and 'model' parameters".format(
                        self.__class__
                    )
                )
            self.ref = "{}_{}".format(inflection.underscore(nested), self.attr)


class HasOne(HasMixin, One):
    def __init__(
        self, nested, attr=None, ref=None, callback=None, callback_kwargs=None, **kwargs
    ):
        """
        HasOne initializer. Uses the "get_one_generic" callback and
        automatically assigns attribute as in the following:

        .. code-block:: python

            model="SampleType", attr="id" # equiv. to 'lambda self: self.sample_type_id.'

        :param nested: model name of the target model
        :type nested: basestring
        :param attr: attribute to append underscored model name
        :type attr: basestring
        """
        self.set_ref(nested=nested, attr=attr, ref=ref)

        super().__init__(
            nested,
            many=False,
            callback=callback,
            callback_args=(self.get_ref,),
            callback_kwargs=callback_kwargs,
            **kwargs
        )

    def get_ref(self, instance):
        return getattr(instance, self.ref)

    def __repr__(self):
        return "<HasOne (model={}, callback_args=lambda self: self.{})>".format(
            self.nested, self.ref
        )

    def deserialize(self, owner, val):
        val = super().deserialize(owner, val)
        if val:
            setattr(owner, self.ref, getattr(val, self.attr))
        return val


class HasMany(HasMixin, Many):
    """
    A relationship that establishes a One-to-Many relationship with another model.
    """

    def __init__(
        self,
        nested,
        ref_model=None,
        attr=None,
        ref=None,
        additional_args=None,
        callback=None,
        callback_kwargs=None,
        **kwargs
    ):
        """
        HasMany relationship initializer

        :param nested: Model class name for this relationship
        :type nested: str
        :param ref_model: Reference model name of the model owning this relationships.

        .. code-block:: python

            @add_schema
            class Author(ModelBase):
                fields=dict(books=HasMany("Book", "Author"))  # search for books using 'author_id'

        :type ref_model: str
        :param attr: Attribute name to use with reference model (default='id').
                     For example "Author" => 'author_id'
        :type attr: str
        :param ref: The reference to use to find models. If none, a reference is
                    built from the 'ref_model' and 'attr' parameters
        :type ref: str
        """
        if ref_model is None and ref is None:
            msg = "'{}' needs a 'ref_model' or 'ref' parameters to initialize"
            raise FieldValidationError(msg.format(self.__class__.__name__))
        self.set_ref(nested=ref_model, attr=attr, ref=ref)

        if additional_args is None:
            additional_args = {}

        def callback_args(slf):
            query = {self.ref: getattr(slf, self.attr)}
            query.update(additional_args)
            return query

        super().__init__(
            nested,
            callback=callback,
            callback_args=callback_args,
            callback_kwargs=callback_kwargs,
            **kwargs
        )


class HasManyThrough(HasMixin, Many):
    """
    A relationship using an intermediate association model.
    Establishes a Many-to-Many relationship with another model
    """

    def __init__(
        self,
        nested,
        through,
        attr="id",
        ref=None,
        additional_args=None,
        callback=None,
        callback_kwargs=None,
        **kwargs
    ):
        self.set_ref(nested=nested, attr=attr, ref=ref)

        # e.g. PlanAssociation >> plan_associations
        through_model_attr = inflection.pluralize(inflection.underscore(through))
        self.through_model_attr = through_model_attr

        if additional_args is None:
            additional_args = {}

        def callback_args(slf):
            through_model = getattr(slf, through_model_attr)
            if through_model is None:
                return None
            query = {
                attr: [getattr(x, self.ref) for x in getattr(slf, through_model_attr)]
            }
            query.update(additional_args)
            return {
                attr: [getattr(x, self.ref) for x in getattr(slf, through_model_attr)]
            }

        super().__init__(
            nested,
            callback=callback,
            callback_args=callback_args,
            callback_kwargs=callback_kwargs,
            **kwargs
        )


class HasOneFromMany(HasMixin, One):
    """
    Returns a single model from a Many relationship
    """

    QUERY_TYPE = "query"

    def __init__(
        self,
        nested,
        ref_model=None,
        attr=None,
        ref=None,
        additional_args=None,
        callback=None,
        callback_kwargs=None,
        **kwargs
    ):
        """
        HasOneFromMany relationship initializer, which is intended to return a single model from
        a Many query

        :param nested: Model class name for this relationship
        :type nested: str
        :param ref_model: Reference model name of the model owning this relationships.

        .. code-block:: python

            @add_schema
            class Author(ModelBase):
                fields=dict(books=HasMany("Book", "Author"))  # search for books using 'author_id'

        :type ref_model: str
        :param attr: Attribute name to use with reference model (default='id').
                     For example "Author" => 'author_id'
        :type attr: str
        :param ref: The reference to use to find models. If none, a reference is
                    built from the 'ref_model' and 'attr' parameters
        :type ref: str
        """
        if ref_model is None and ref is None:
            msg = "'{}' needs a 'ref_model' or 'ref' parameters to initialize"
            raise FieldValidationError(msg.format(self.__class__.__name__))
        self.set_ref(nested=ref_model, attr=attr, ref=ref)

        if additional_args is None:
            additional_args = {}

        def callback_args(slf):
            query = {self.ref: getattr(slf, self.attr)}
            query.update(additional_args)
            return query

        if callback is None:
            callback = ModelBase.one_callback.__name__

        super().__init__(
            nested,
            callback=callback,
            callback_args=callback_args,
            callback_kwargs=callback_kwargs,
            **kwargs
        )


class HasManyGeneric(HasMany):
    """
    Establishes a One-to-Many relationship using 'parent_id' as the attribute
    to find other models.
    """

    def __init__(
        self,
        nested,
        additional_args=None,
        callback=None,
        callback_kwargs=None,
        **kwargs
    ):
        super().__init__(
            nested,
            ref="parent_id",
            attr="id",
            callback=callback,
            additional_args=additional_args,
            callback_kwargs=callback_kwargs,
            **kwargs
        )



"""
Models related to jobs, or operation execution.
"""

from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.relationships import JSON, HasOne, HasMany, HasManyThrough


@add_schema
class Job(ModelBase):
    """A Job model"""

    fields = dict(
        job_associations=HasMany("JobAssociation", "Job"),
        operations=HasManyThrough("Operation", "JobAssociation"),
        state=JSON(),
    )

    @property
    def is_complete(self):
        return self.pc == -2

    @property
    def uploads(self):
        http = self.session._AqSession__aqhttp
        return http.get("krill/uploads?job={}".format(self.id))["uploads"]

    @property
    def start_time(self):
        return self.state[0]["time"]

    @property
    def end_time(self):
        return self.state[-2]["time"]


@add_schema
class JobAssociation(ModelBase):
    """A JobAssociation model"""

    fields = dict(job=HasOne("Job"), operation=HasOne("Operation"))



def make_signature_str(_args, _kwargs):
    return "({}, {})".format(
        ", ".join([str(_a) for _a in _args]),
        ", ".join(["{}={}".format(name, val) for name, val in _kwargs.items()]),
    )



"""
Models related to plans, plan associations, and wires.
"""

import json

from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.marshaller import add_schema
from pydent.models.field_value import FieldValue
from pydent.models.data_associations import Upload, DataAssociatorMixin
from pydent.relationships import (
    Raw,
    JSON,
    Many,
    HasOne,
    HasMany,
    HasManyThrough,
    HasManyGeneric,
    fields,
)
from pydent.models.crud_mixin import SaveMixin, DeleteMixin
from warnings import warn


@add_schema
class Plan(DataAssociatorMixin, SaveMixin, DeleteMixin, ModelBase):
    """
    A Plan model
    """

    fields = dict(
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Plan"}
        ),
        plan_associations=HasMany("PlanAssociation", "Plan"),
        operations=HasManyThrough("Operation", "PlanAssociation"),
        wires=Many("Wire", callback="_get_wires_from_server"),
        layout=JSON(),
        status=Raw(default="planning"),
    )
    query_hook = {"include": "wires"}

    def __init__(self, name="MyPlan", status="planning"):
        super().__init__(
            name=name,
            status=status,
            data_associations=None,
            plan_associations=None,
            layout={
                "id": 0,
                "children": [],
                "documentation": "No documentation of this module",
                "height": 60,
                "input": [],
                "output": [],
                "name": "no_name",
                "parent_id": -1,
                "width": 160,
                "wires": [],
            },
        )

    def add_operation(self, operation):
        """
        Adds an operation to the Plan

        :param operation: Operation to add
        :type operation: Operation
        :return: None
        :rtype: None
        """
        self.append_to_many("operations", operation)

    def add_operations(self, operations):
        """
        Adds multiple operations to the Plan

        :param operations: list of Operations
        :type operations: list
        :return: None
        :rtype: None
        """
        for operation in operations:
            self.add_operation(operation)

    def has_operation(self, op):
        if op is None:
            return False
        return self.operations and op.rid in [_op.rid for _op in self.operations]

    def find_wires(self, src, dest):
        """
        Retrieves the wire between a source and destination FieldValues

        :param src: source FieldValue
        :type src: FieldValue
        :param dest: destination FieldValue
        :type dest: FieldValue
        :return: array of wires between src and dest FieldValues (determined by rid)
        :rtype: array of wires
        """

        found = []
        for wire in self.wires:
            if src.rid == wire.source.rid and dest.rid == wire.destination.rid:
                found.append(wire)
        return found

    def wire(self, src, dest):
        """
        Creates a new wire between src and dest FieldValues. Returns the new wire
        if it does not exist in the plan. If the wire already exists and
        error_if_exists is True, then the existing wire is returned. Else an exception is raised.

        :param src: source field value (the input of the wire)
        :type src: FieldValue
        :param dest: destination field value (the output of the wire)
        :type dest: FieldValue
        :param error_if_exists: Raise an error if the Wire already exists in the plan.
        :type error_if_exists: boolean
        :return: Newly created wire or existing wire (if exists and error_if_exists == False)
        :rtype: Wire
        """

        if not self.has_operation(src.operation):
            raise AquariumModelError(
                "Cannot wire because the wire's source FieldValue {} does "
                "not exist in the Plan because its Operation '{}' is not in the plan".format(
                    src, src.operation
                )
            )
        if not self.has_operation(dest.operation):
            raise AquariumModelError(
                "Cannot wire because the wire's destination FieldValue {} does not exist "
                "in the Plan because its Operation '{}' is not in the plan.".formst(
                    dest, dest.operation
                )
            )

        wire = Wire(source=src, destination=dest)
        self.append_to_many("wires", wire)
        return wire

    def _collect_wires(self):
        incoming_wires = []
        outgoing_wires = []
        if self.operations:
            for operation in self.operations:
                if operation.field_values:
                    for field_value in operation.field_values:
                        if field_value.outgoing_wires:
                            outgoing_wires += field_value.outgoing_wires
                        if field_value.incoming_wires:
                            incoming_wires += field_value.incoming_wires
        return incoming_wires, outgoing_wires

    def _get_wire_dict(self, wires):
        """Return all wires in the plan grouped by the wire identifier"""
        wire_dict = {}
        for w in wires:
            wire_dict.setdefault(w.identifier, list())
            if w not in wire_dict[w.identifier]:
                wire_dict[w.identifier].append(w)
        return wire_dict

    def _get_wires_from_server(self, *args):
        fvs = []
        if self.operations:
            for op in self.operations:
                fvs += op.field_values

        fv_ids = [fv.id for fv in fvs if fv.id is not None]
        wires_from_server = []
        if fv_ids:
            wires_from_server = self.session.Wire.where(
                {"from_id": fv_ids, "to_id": fv_ids}
            )
        return wires_from_server

    def submit(self, user, budget):
        """
        Submits the Plan to the Aquarium server.

        :param user: User to submit the Plan
        :type user: User
        :param budget: Budget to use for the Plan
        :type budget: Budget
        :return: JSON
        :rtype: dict
        """
        result = self.session.utils.submit_plan(self, user, budget)
        return result

    def all_data_associations(self):
        das = self.data_associations
        for operation in self.operations:
            das += operation.data_associations
            for field_value in operation.field_values:
                if field_value.item:
                    das += field_value.item.data_associations
        return das

    @classmethod
    def interface(cls, session):
        # get model interface from Base class
        model_interface = super(Plan, cls).interface(session)

        # make a special find method for plans, as generic method is too minimal.
        def new_find(model_id):
            return model_interface.get("plans/{}.json".format(model_id))

        # override the old find method
        model_interface.find = new_find

        return model_interface

    def validate(self, raise_error=True):
        """
        Validates the plan.

        :param raise_error: If True, raises an AquariumModelException. If false, returns the error messages.
        :type raise_error: boolean
        :return: list of error messages
        :rtype: array
        """
        errors = []

        field_values = []
        for op in self.operations:
            for fv in op.field_values:
                field_values.append(fv)
        fv_keys = [fv._primary_key for fv in field_values]
        for wire in self.wires:
            for _fvtype in ["source", "destination"]:
                field_value = getattr(wire, _fvtype)
                if field_value._primary_key not in fv_keys:
                    msg = (
                        "The FieldValue of a wire Wire(rid={}).{} is missing from the list"
                        " of FieldValues in the plan. Did you forget to add an operation to "
                        " the plan?".format(wire._primary_key, _fvtype)
                    )
                    errors.append(msg)
        if raise_error and errors:
            msg = "\n".join(
                ["(ErrNo {}) - {}".format(i, e) for i, e in enumerate(errors)]
            )
            raise AquariumModelError(
                "Plan {} had the following errors:\n{}".format(self, msg)
            )
        return errors

    def to_save_json(self):
        """
        Returns the json representation of the plan for saving and creating
        Plans on the Aquarium server.

        :return: JSON
        :rtype: dict
        """
        if not self.operations:
            self.operations = []

        if not self.wires:
            self.wires = []

        self.validate(raise_error=True)

        for op in self.operations:
            op.field_values

        json_data = self.dump(include={"operations": {"field_values": {}}})

        # remove redundant wires
        wire_dict = {}
        for wire in self.wires:
            wire_data = wire.to_save_json()
            attributes = [
                wire_data["from_id"],
                wire_data["from"]["rid"],
                wire_data["to_id"],
                wire_data["to"]["rid"],
            ]
            wire_hash = "*&".join([str(a) for a in attributes])
            wire_dict[wire_hash] = wire_data
        json_data["wires"] = list(wire_dict.values())

        # validate
        fv_rids = []
        fv_id_to_rids = {}
        for op in json_data["operations"]:
            for fv in op["field_values"]:
                if fv["id"]:
                    fv_id_to_rids[fv["id"]] = fv["rid"]
                fv_rids.append(fv["rid"])

        # fix json rids and ids
        warnings = []
        for wire_data in json_data["wires"]:
            if wire_data["from_id"] in fv_id_to_rids:
                wire_data["from"]["rid"] = fv_id_to_rids[wire_data["from_id"]]
            if wire_data["to_id"] in fv_id_to_rids:
                wire_data["to"]["rid"] = fv_id_to_rids[wire_data["to_id"]]
            if not wire_data["from"]["rid"] in fv_rids:
                warnings.append(
                    "FieldValue rid={} is missing.".format(wire_data["from"]["rid"])
                )
            if not wire_data["to"]["rid"] in fv_rids:
                warnings.append(
                    "FieldValue rid={} is missing.".format(wire_data["to"]["rid"])
                )
        if warnings:
            print(fv_rids)
            warn(",".join(warnings))

        if json_data["layout"] is not None:
            json_data["layout"] = json.loads(json_data["layout"])
        else:
            del json_data["layout"]

        return json_data

    def _get_create_json(self):
        return self.to_save_json()

    def _get_update_json(self):
        return self.to_save_json()

    def _get_create_params(self):
        return {"user_id": self.session.current_user.id}

    def _get_update_params(self):
        return {"user_id": self.session.current_user.id}

    def estimate_cost(self):
        """
        Estimates the cost of the plan on the Aquarium server.
        This is necessary before plan submission.

        :return: cost
        :rtype: dict
        """
        return self.session.utils.estimate_plan_cost(self)

    def field_values(self):
        raise NotImplementedError()

    def step(self):
        """Steps a plan"""
        return self.session.utils.step_plan(self.id)

    def show(self):
        """Print the plan nicely"""
        print(self.name + " id: " + str(self.id))
        for operation in self.operations:
            operation.show(pre="  ")
        for wire in self.wires:
            wire.show(pre="  ")

    def replan(self):
        """Copies or replans the plan. Returns a plan copy"""
        return self.session.utils.replan(self.id)

    def download_files(self, outdir=None, overwrite=True):
        """
        Downloads all uploads associated with the plan. Downloads happen
        ansynchrounously.

        :param outdir: output directory for downloaded files
        :param overwrite: whether to overwrite files if they exist
        :return: None
        """
        uploads = []
        for da in self.data_associations:
            if da.upload is not None:
                uploads.append(da.upload)
        return Upload.async_download(uploads, outdir, overwrite)


@add_schema
class PlanAssociation(ModelBase):
    """A PlanAssociation model"""

    fields = dict(plan=HasOne("Plan"), operation=HasOne("Operation"))

    def __init__(self, plan_id=None, operation_id=None):
        super().__init__(plan_id=plan_id, operation_id=operation_id)


@add_schema
class Wire(DeleteMixin, ModelBase):
    """A Wire model"""

    fields = {
        "source": HasOne("FieldValue", ref="from_id"),
        "destination": HasOne("FieldValue", ref="to_id"),
    }

    WIRABLE_PARENT_CLASSES = ["Operation"]

    def __init__(self, source=None, destination=None):

        self._validate_field_values(source, destination)

        if hasattr(source, "id"):
            from_id = source.id
        else:
            from_id = None

        if hasattr(destination, "id"):
            to_id = destination.id
        else:
            to_id = None

        if (source and not destination) or (destination and not source):
            raise AquariumModelError(
                "Cannot wire. Either source ({}) or destination ({}) is None".format(
                    source, destination
                )
            )

        if source:
            if not source.role:
                raise AquariumModelError(
                    "Cannot wire. FieldValue {} does not have a role".format(
                        source.role
                    )
                )
            elif source.role != "output":
                raise AquariumModelError(
                    "Cannot wire an '{}' FieldValue as a source".format(source.role)
                )

        if destination:
            if not destination.role:
                raise AquariumModelError(
                    "Cannot wire. FieldValue {} does not have a role".format(
                        destination.role
                    )
                )
            elif destination.role != "input":
                raise AquariumModelError(
                    "Cannot wire an '{}' FieldValue as a destination".format(
                        destination.role
                    )
                )

        super().__init__(
            **{
                "source": source,
                "from_id": from_id,
                "destination": destination,
                "to_id": to_id,
            },
            active=True
        )

    @property
    def identifier(self):

        if not self.source:
            source_id = self.from_id
        else:
            source_id = "r" + str(self.source.rid)

        if not self.destination:
            destination_id = self.to_id
        else:
            destination_id = "r" + str(self.destination.rid)
        return "{}_{}".format(source_id, destination_id)

    @classmethod
    def _validate_field_values(cls, src, dest):
        if src:
            cls._validate_field_value(src, "source")
        if dest:
            cls._validate_field_value(dest, "destination")

        if src and dest and src.rid == dest.rid:
            raise AquariumModelError(
                "Cannot create wire because source and destination are the same instance."
            )

    @classmethod
    def _validate_field_value(cls, fv, name):
        if not issubclass(type(fv), FieldValue):
            raise AquariumModelError(
                "Cannot create wire because {} FieldValue is {}.".format(
                    name, fv.__class__.__name__
                )
            )

        if fv.parent_class not in cls.WIRABLE_PARENT_CLASSES:
            raise AquariumModelError(
                "Cannot create wire because the {} FieldValue is has '{}' parent class. "
                "Only {} parent classes are wirable".format(
                    name, fv.parent_class, cls.WIRABLE_PARENT_CLASSES
                )
            )

    def validate(self):
        self._validate_field_values(self.source, self.destination)

    def to_save_json(self):
        save_json = {
            "id": self.id,
            "from_id": self.source.id,
            "to_id": self.destination.id,
            "from": {"rid": self.source.rid},
            "to": {"rid": self.destination.rid},
            "active": self.active,
        }
        return save_json

    def delete(self):
        """
        Permanently deletes the wire instance on the Aquarium server.
        :return:
        :rtype:
        """
        return self.session.utils.delete_wire(self)

    def show(self, pre=""):
        """Show the wire nicely"""
        source = self.source
        dest = self.destination

        from_op_type_name = "None"
        from_name = "None"
        if source:
            from_op_type_name = self.source.operation.operation_type.name
            from_name = source.name

        to_op_type_name = "None"
        to_name = "None"
        if dest:
            to_op_type_name = self.destination.operation.operation_type.name
            to_name = dest.name

        print(
            pre
            + from_op_type_name
            + ":"
            + from_name
            + " --> "
            + to_op_type_name
            + ":"
            + to_name
        )

    def does_wire_to(self, destination):
        if destination and self.destination:
            return destination.rid == self.destination.rid
        return False

    def does_wire_from(self, source):
        if source and self.source:
            return source.rid == self.source.rid
        return False

    def does_wire(self, source, destination):
        """Checks whether this Wire is a wire between the source and destination FieldValues.

        If any of the source or destination FieldValues are None, returns False."""
        if source and destination and self.source and self.destination:
            return (
                source.rid == self.source.rid
                and destination.rid == self.destination.rid
            )
        return False

    def __eq__(self, other):
        """Checks whether this Wire is wired to the same FieldValue instances of another Wire.

        see `does_wire` method."""
        return self.does_wire(other.source, other.destination)



# __version__.py
# autogenerated by keats 0.2.25
__version__ = "0.1.5a6"
__name__ = "pydent"
__title__ = "pydent"
__authors__ = [
    "jvrana <justin.vrana@gmail.com>",
    "Ben Keller <bjkeller@uw.edu>",
    "Eric Klavins <klavins@uw.edu>",
]
__repo__ = None
__homepage__ = "https://www.github.com/klavinslab/trident"
__description__ = "Aquarium's Python API for planning, executing, and analyzing scientific experiments."



"""
Model base class
"""

from pydent.marshaller.exceptions import SchemaException, SchemaModelException
from pydent.marshaller.schema import DynamicSchema, SchemaRegistry
from pydent.marshaller.fields import Callback
from pydent.marshaller.registry import ModelRegistry
from pydent.marshaller.descriptors import DataAccessor
from copy import deepcopy
import functools
import inspect


def add_schema(cls):
    """Decorator that dynamically attaches a schema to a model."""

    if not issubclass(cls, SchemaModel):
        raise SchemaException("Model must be a subclaass of '{}'".format(SchemaModel))
    new_schema = SchemaRegistry(
        SchemaRegistry.make_schema_name(cls.__name__),
        (DynamicSchema,),
        {"_model_class": cls},
    )
    new_schema.register(cls)
    return cls


class SchemaModel(metaclass=ModelRegistry):
    """
    A Schema class that holds information on serialization/deserialization
    """

    __slots__ = [ModelRegistry._data_key, ModelRegistry._deserialized_key, "__dict__"]

    def __init__(self, data=None):
        """
        The model initializer.

        :param data: data to add to the model
        :type data: dict
        """
        if data is None:
            data = {}
        if not getattr(self, ModelRegistry._data_key, None):
            setattr(self, ModelRegistry._data_key, data)
        if not getattr(self, ModelRegistry._deserialized_key, None):
            setattr(self, ModelRegistry._deserialized_key, {})
        self.add_data(data)

    @property
    def model_schema(self):
        """
        Returns the attached model schema.

        :return: the model's schema class
        :rtype: DynamicSchema
        """
        return getattr(self.__class__, "model_schema", None)

    def add_data(self, data):
        """Initializes fake attributes that correspond to data."""
        if not self.model_schema:
            filepath = ""
            try:
                filepath = inspect.getfile(self.__class__)
            except:
                pass
            raise SchemaModelException(
                "Cannot initialize a {} without a {}. "
                "Use '@{}' to decorate the class definition for '{}' located in {}".format(
                    SchemaModel.__name__,
                    DynamicSchema.__name__,
                    add_schema.__name__,
                    self.__class__.__name__,
                    filepath,
                )
            )
        if data is not None:
            self.__class__.model_schema.init_data_accessors(self, data)

    def _get_data(self):
        """Return the model's data"""
        return getattr(self, self.__class__._data_key)

    def _get_deserialized_data(self):
        return getattr(self, self.__class__._deserialized_key)

    def get_deserialized(self, name):
        return self._get_deserialized_data()[name]

    def has_deserialize(self, name):
        assert name in self._get_deserialized_data()

    @classmethod
    def _set_data(cls, data, calling_obj=None):
        instance = SchemaModel.__new__(cls)
        SchemaModel.__init__(instance, data)
        return instance

    @staticmethod
    def _keys_to_dict(keys):
        if keys is None:
            return {}
        elif isinstance(keys, str):
            return {keys: None}
        elif isinstance(keys, dict):
            return dict(keys)
        else:
            return {k: None for k in keys}

    @classmethod
    def _dump(cls, obj, only=None, include=None, ignore=None):
        if not issubclass(type(obj), SchemaModel):
            return obj

        only = cls._keys_to_dict(only)
        include = cls._keys_to_dict(include)
        ignore = cls._keys_to_dict(ignore)

        data = {}

        serialized_data = obj._get_data()

        include_and_only = set(include).union(set(only))

        for fname in include_and_only:
            getattr(obj, fname)

        keys = set(serialized_data.keys())
        keys = keys.difference(set(ignore))
        if only:
            keys = keys.intersection(set(only))

        model_fields = obj.__class__.model_schema.fields

        callback_fields = []
        always_dump = []
        for fname, field in model_fields.items():
            if hasattr(field, "nested") or issubclass(type(field), Callback):
                callback_fields.append(fname)
            if getattr(field, "always_dump", None):
                always_dump.append(fname)

        include_and_only = include_and_only.union(set(always_dump))
        callback_keys = include_and_only.intersection(set(callback_fields))
        callback_keys = callback_keys.difference(set(ignore))

        data_keys = keys.difference(set(callback_fields))

        for key in data_keys:
            data[key] = serialized_data[key]
        for key in callback_keys:
            field = model_fields[key]
            val = getattr(obj, key)
            dump = functools.partial(
                cls._dump,
                include=include.pop(key, None),
                only=only.pop(key, None),
                ignore=ignore.pop(key, None),
            )
            if getattr(model_fields[key], "nested", None) and model_fields[key].many:
                if isinstance(val, list):
                    data[field.data_key] = [dump(v) for v in val]
            else:
                data[field.data_key] = dump(val)
        return data

    def dump(self, only=None, include=None, ignore=None):
        """
        Dump/serializes the model to a json-like dictionary.

        :param only: restricts dump/serialization to the provided keys
        :type only: basestring|list|tuple|dict
        :param include: include any callback fields in the dump/serialization
        :type include: basestring|list|tuple|dict
        :param ignore: ignores fields in the dump/serialization
        :type ignore: basestring|list|tuple|dict
        :return: the serialized data
        :rtype: dict
        """
        return deepcopy(self._dump(self, only=only, include=include, ignore=ignore))

    # def __dir__(self):
    #     return ['a']

    @classmethod
    def load(cls, data):
        """
        Deserialize the data to a model.

        :param data: data to deserialize
        :type data: dict
        :return: the deserialized model
        :rtype: SchemaModel
        """
        return cls._set_data(data)

    def __setstate__(self, state):
        """Overrides how models are unpickled. The default way pickling works by dumping
        the __dict__ along with the class definition. Since the way data is handle in
        the marshaller object involves dynamic creation of descriptors, these are not properly
        handled upon load. Here we will create these descriptors upon
        unpickling. Unlike serialized data, deserialized data is handled explicitly using 'fields' in the
        schema/model definition, it is not necessary to do anything with these descriptors."""
        self.__dict__ = state

        # add data accessors
        data = state[ModelRegistry._data_key]
        for k in data:
            if k not in self.__class__.__dict__:
                setattr(self.__class__, k, DataAccessor(k, ModelRegistry._data_key))



"""
Models related to operations and operation types.
"""

from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.marshaller import add_schema
from pydent.models.data_associations import DataAssociatorMixin
from pydent.models.field_value_mixins import FieldValueInterface, FieldTypeInterface
from pydent.relationships import (
    Raw,
    Function,
    HasOne,
    HasMany,
    HasManyThrough,
    HasManyGeneric,
    HasOneFromMany,
)
from pydent.utils import filter_list
from pydent.models.crud_mixin import SaveMixin


@add_schema
class Operation(FieldValueInterface, DataAssociatorMixin, ModelBase):
    """A Operation model"""

    fields = dict(
        field_values=HasMany(
            "FieldValue", ref="parent_id", additional_args={"parent_class": "Operation"}
        ),
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Operation"}
        ),
        operation_type=HasOne("OperationType"),
        job_associations=HasMany("JobAssociation", "Operation"),
        jobs=HasManyThrough("Job", "JobAssociation"),
        plan_associations=HasMany("PlanAssociation", "Operation"),
        plans=HasManyThrough("Plan", "PlanAssociation"),
        status=Raw(default="planning"),
        routing=Function("get_routing"),
    )

    METATYPE = "operation_type"

    def __init__(
        self, operation_type_id=None, operation_type=None, status=None, x=0, y=0
    ):
        super().__init__(
            operation_type_id=operation_type_id,
            operation_type=operation_type,
            status=status,
            field_values=None,
            x=x,
            y=y,
        )

    def get_routing(self):
        routing_dict = {}
        fvs = self.field_values
        ot = self.operation_type
        if ot is None:
            return routing_dict
        for ft in ot.field_types:
            if ft.routing is not None:
                routing_dict[ft.routing] = None
        if fvs is not None:
            for fv in self.field_values:
                ft = self.safe_get_field_type(fv)
                if ft.routing is not None:
                    routing_dict[ft.routing] = fv.sid
        return routing_dict

    @property
    def successors(self):
        successors = []
        if self.outputs:
            for output in self.outputs:
                for s in output.successors:
                    successors.append(s.operation)
        return successors

    @property
    def predecessors(self):
        predecessors = []
        if self.inputs:
            for inputs in self.inputs:
                for s in inputs.predecessors:
                    predecessors.append(s.operation)
        return predecessors

    def init_field_values(self):
        """
        Initialize the :class:`FieldValue` from the :class:`FieldType` of the
        parent :class:`Operation` type.
        """
        self.field_values = []
        for field_type in self.get_metatype().field_types:
            if not field_type.array:
                self.new_field_value_from_field_type(field_type)
        return self

    def field_value_array(self, name, role):
        """Returns :class:`FieldValue` array with name and role."""
        return filter_list(self.get_field_value_array(name, role))

    def field_value(self, name, role):
        """
        Returns :class:`FieldValue` with name and role.
        Return None if not found.
        """
        if self.field_values:
            fvs = self.field_value_array(name, role)

            if len(fvs) == 0:
                return None

            if len(fvs) == 1:
                return fvs[0]

            msg = "More than one FieldValue found for the field value"
            msg += " of operation {}.(id={}).{}.{}. Are you sure you didn't mean to call 'field_value_array'?"
            raise AquariumModelError(
                msg.format(self.operation_type, self.id, role, name)
            )

    @property
    def plan(self):
        return self.plans[0]

    def input_array(self, name):
        return self.get_field_value_array(name, "input")

    def output_array(self, name):
        return self.get_field_value_array(name, "output")

    def input(self, name):
        """Returns the input :class:`FieldValue` by name"""
        return self.field_value(name, "input")

    def output(self, name):
        """Returns the output :class:`FieldValue` by name"""
        return self.field_value(name, "output")

    def add_to_input_array(
        self, name, sample=None, item=None, value=None, container=None
    ):
        """
        Creates and adds a new input :class:`FieldValue`.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValue
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the newly created FieldValue
        :rtype: FieldValue
        """
        return self.new_field_value(
            name,
            "input",
            dict(sample=sample, item=item, value=value, container=container),
        )

    def add_to_output_array(
        self, name, sample=None, item=None, value=None, container=None
    ):
        """
        Creates and adds a new output :class:`FieldValue`.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValue
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the newly created FieldValue
        :rtype: FieldValue
        """
        return self.new_field_value(
            name,
            "output",
            dict(sample=sample, item=item, value=value, container=container),
        )

    @property
    def inputs(self):
        """Return a list of all input :class:`FieldValues`"""
        return [fv for fv in self.field_values if fv.role == "input"]

    @property
    def outputs(self):
        """Return a list of all output :class:`FieldValues`"""
        return [fv for fv in self.field_values if fv.role == "output"]

    def set_input(
        self, name, sample=None, item=None, value=None, container=None, object_type=None
    ):
        """
        Sets a input :class:`FieldValue` to a value. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldValue/FieldType
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the existing FieldValue modified
        :rtype: FieldValue
        """
        if object_type is None and container:
            object_type = container
        return self.set_field_value(
            name,
            "input",
            dict(sample=sample, item=item, value=value, object_type=container),
        )

    def set_output(
        self, name, sample=None, item=None, value=None, container=None, object_type=None
    ):
        """
        Sets a output :class:`FieldValue` to a value. When setting values to
        items/samples/containers, the item/sample/container must be saved.

        :param name: name of the FieldValue/FieldType
        :type name: string
        :param sample: an existing Sample
        :type sample: Sample
        :param item: an existing Item
        :type item: Item
        :param value: a string or number value
        :type value: string|integer
        :param container: an existing ObjectType
        :type container: ObjectType
        :return: the existing FieldValue modified
        :rtype: FieldValue
        """
        if object_type is None and container:
            object_type = container
        return self.set_field_value(
            name,
            "output",
            dict(sample=sample, item=item, value=value, object_type=object_type),
        )

    def set_input_array(self, name, values):
        """
        Sets input :class:`FieldValue` array using values.
        Values should be a list of dictionaries containing sample, item,
        container, or values keys.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValues being modified
        :type name: string
        :param values: list of dictionary of values to set
                (e.g. [{"sample": mysample}, {"item": myitem}])
        :type values: list
        :return: the list of modified FieldValues
        :rtype: list
        """
        return self.set_field_value_array(name, "input", values)

    def set_output_array(self, name, values):
        """
        Sets output :class:`FieldValue` array using values.
        Values should be a list of dictionaries containing sample, item,
        container, or values keys.
        When setting values to items/samples/containers, the
        item/sample/container must be saved.

        :param name: name of the FieldType/FieldValues being modified
        :type name: string
        :param values: list of dictionary of values to set
                (e.g. [{"sample": mysample}, {"item": myitem}])
        :type values: list
        :return: the list of modified FieldValues
        :rtype: list
        """
        return self.set_field_value_array(name, "output", values)

    def show(self, pre=""):
        """
        Print the operation nicely
        """
        print(pre + self.operation_type.name + " " + str(self.cost))
        for field_value in self.field_values:
            field_value.show(pre=pre + "  ")

    def __str__(self):
        return self._to_str(operation_type_name=self.operation_type.name)


@add_schema
class OperationType(FieldTypeInterface, SaveMixin, ModelBase):
    """
    Represents an OperationType, which is the definition of a protocol in
    Aquarium.
    """

    fields = dict(
        operations=HasMany("Operation", "OperationType"),
        field_types=HasMany(
            "FieldType",
            ref="parent_id",
            additional_args={"parent_class": "OperationType"},
        ),
        codes=HasManyGeneric("Code"),
        cost_model=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "cost_model"},
        ),
        documentation=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "documentation"},
        ),
        precondition=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "precondition"},
        ),
        protocol=HasOneFromMany(
            "Code",
            ref="parent_id",
            additional_args={"parent_class": "OperationType", "name": "protocol"},
        ),
    )

    def code(self, accessor):
        # raise DeprecationWarning("This method is depreciated. Call code directly using 'protocol', 'cost_model',"
        #                          " 'documentation', or 'precondition'.")
        if accessor in ["protocol", "precondition", "documentation", "cost_model"]:
            return getattr(self, accessor)
        return None

    def instance(self, xpos=0, ypos=0):
        operation = self.session.Operation.new(
            operation_type_id=self.id, status="planning", x=xpos, y=ypos
        )
        operation.operation_type = self
        operation.init_field_values()
        return operation

    def field_type(self, name, role):
        if self.field_types:
            fts = filter_list(self.field_types, role=role, name=name)
            if len(fts) > 0:
                return fts[0]

    def output(self, name):
        return self.field_type(name, "output")

    def input(self, name):
        return self.field_type(name, "input")

    def save(self):
        """Saves the Operation Type to the Aquarium server. Requires
        this Operation Type to be connected to a session."""
        return self.reload(self.session.utils.create_operation_type(self))

    def to_save_json(self):
        op_data = self.dump(
            include={
                "field_types": {"allowable_field_types": {}},
                "protocol": {},
                "cost_model": {},
                "documentation": {},
                "precondition": {},
            }
        )

        # Format 'sample_type' and 'object_type' keys for afts
        for ft_d, ft in zip(op_data["field_types"], self.field_types):
            for aft_d, aft in zip(
                ft_d["allowable_field_types"], ft.allowable_field_types
            ):
                aft_d["sample_type"] = {"name": aft.sample_type.name}
                aft_d["object_type"] = {"name": aft.object_type.name}
        return op_data

    def _get_create_json(self):
        return self.to_save_json()

    def _get_update_json(self):
        return self.to_save_json()

    def __str__(self):
        return self._to_str("id", "name", "category")



"""
The create, read, update, destroy (CRUD) mixins used for some models.
"""


class CreateMixin(object):
    def create(self):
        data = self._get_create_json()
        params = self._get_create_params()
        name = self.get_tableized_name()
        result = self.session.utils.model_create(name, data, params=params)
        self.reload(result)
        return self

    def _get_create_json(self):
        return self.dump()

    def _get_create_params(self):
        return None


class UpdateMixin(object):
    def update(self):
        data = self._get_create_json()
        params = self._get_update_params()
        name = self.get_tableized_name()
        result = self.session.utils.model_update(name, self.id, data, params=params)
        self.reload(result)
        return self

    def _get_update_json(self):
        return self.dump()

    def _get_update_params(self):
        return None


class SaveMixin(UpdateMixin, CreateMixin):
    def save(self):
        if self.id:
            self.update()
        else:
            self.create()


class DeleteMixin(object):
    def delete(self):
        name = self.get_tableized_name()
        params = self._get_delete_params()
        result = self.session.utils.model_delete(name, self.id, params=params)
        return result

    def _get_delete_params(self):
        return None


class JSONSaveMixin(object):
    def save(self):
        name = self.get_server_model_name()
        data = self._get_save_json()
        result = self.session.utils.json_save(name, data)
        return result

    def _get_save_json(self):
        return self.dump()


class JSONDeleteMixin(object):
    def delete(self):
        name = self.get_server_model_name()
        data = self._get_delete_json()
        return self.session.utils.json_delete(name, data)

    def _get_delete_json(self):
        return self.dump()



"""
User, group, and budget models.
"""

from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.relationships import HasOne, HasMany, HasManyThrough


@add_schema
class User(ModelBase):
    """A User model"""

    fields = dict(
        groups=HasMany("Group", "User"),
        user_budget_associations=HasMany("UserBudgetAssociation", "User"),
        budgets=HasManyThrough("Budget", "UserBudgetAssociation"),
        additional=("name", "id", "login"),
        ignore=("password_digest", "remember_token", "key"),
    )

    def __init__(self):
        pass


@add_schema
class UserBudgetAssociation(ModelBase):
    """An association model between a User and a Budget"""

    fields = dict(budget=HasOne("Budget"), user=HasOne("User"))

    def __init__(self):
        pass


@add_schema
class Membership(ModelBase):
    fields = dict(user=HasOne("User"), group=HasOne("Group"))


@add_schema
class Group(ModelBase):
    """A Group model"""

    pass


@add_schema
class Account(ModelBase):
    """An Account model"""


@add_schema
class Budget(ModelBase):
    """A Budget model"""

    fields = dict(user_budget_associations=HasMany("UserBudgetAssociation", "Budget"))


@add_schema
class Invoice(ModelBase):
    """A Invoice model"""

    pass



"""
Request class for making raw http requests to Aquarium

This module contains the AqHTTP class, which can make arbitrary post/put/get
requests to Aquarium and returns JSON data.

Generally, Trident users should be unable to make arbitrary requests using this
class.
Users should only access these methods indirectly through a ``Session`` or
``SessionInterface`` instance.
"""

import json

import requests

from pydent.exceptions import (
    TridentRequestError,
    TridentLoginError,
    TridentTimeoutError,
    TridentJSONDataIncomplete,
    ForbiddenRequestError,
)
from pydent.utils import url_build, Loggable


class AqHTTP(object):
    """
    Defines a session/connection to Aquarium.
    Makes HTTP requests to Aquarium and returns JSON.

    This class should be obscured from Trident user so that users cannot make
    arbitrary requests to an Aquarium server and get sensitive information
    (e.g. User json that is returned contains api_key, password_digest, etc.)
    or make damaging posts.
    Instead, a SessionInterface should be the object that makes these requests.
    """

    TIMEOUT = 10

    def __init__(self, login, password, aquarium_url):
        """
        Initializes an aquarium session with login, password, and server.

        :param login: Aquarium login
        :type login: str
        :param aquarium_url: aquarium url to the server
        :type aquarium_url: str
        """
        self.login = login
        self.aquarium_url = aquarium_url
        self._requests_session = None
        self.timeout = self.__class__.TIMEOUT
        self._login(login, password)
        self.log = Loggable(self, name="AqHTTP@{}".format(aquarium_url))
        self._using_requests = True
        self.num_requests = 0

    def on(self):
        self._using_requests = True

    def off(self):
        self._using_requests = False

    def _format_response_info(self, response, include_text=False, include_body=True):
        if response is not None:
            if response.status_code >= 400:
                include_text = True
            info = dict(response.request.__dict__)
            info.update({"seconds": response.elapsed.total_seconds()})
            msg = "REQUEST: (t={seconds}s)  {method} {url}".format(**info)
            if include_body:
                body = info["body"]
                try:
                    my_json = body.decode("utf8").replace("'", '"')
                    body = json.loads(my_json)
                    body = self._pprint_data(body, max_list_len=10)
                except:
                    pass
                msg = msg + "\n" + "BODY: {body}".format(body=body)
            if include_text:
                text = getattr(response, "text", "")
                try:
                    text = json.dumps(json.loads(text), indent=2)
                except:
                    pass
                msg = msg + "\n" + "TEXT: {text}".format(text=text)
            return msg
        return "RESPONSE: NO RESPONSE"

    def _format_request_status(self, response):
        if response is not None:
            return "STATUS:  {} {}".format(response.status_code, response.reason)
        return "STATUS: NO RESPONSE"

    @property
    def url(self):
        """An alias of aquarium_url"""
        return self.aquarium_url

    @staticmethod
    def create_session_json(login, password):
        return {"session": {"login": login, "password": password}}

    def _login(self, login, password):
        """
        Login to aquarium and saves header as a requests.Session()
        """
        session_data = self.create_session_json(login, password)
        try:
            res = requests.post(
                url_build(self.aquarium_url, "sessions.json"),
                json=session_data,
                timeout=self.timeout,
            )
            if res is None:
                raise requests.exceptions.ConnectionError
            cookies = dict(res.cookies)

            # Check for remember token
            if not any(["remember_token" in k for k in dict(res.cookies)]):
                raise TridentLoginError(
                    "Authentication error. Remember token not found in login request."
                    " Contact developers."
                )

            # fix remember token (for some outdated versions of Aquarium)
            for c in dict(cookies):
                if "remember_token" in c:
                    cookies["remember_token"] = cookies[c]
            # TODO: do we remove the session cookie to handle asynchronous requests?
            self.cookies = dict(cookies)
        except requests.exceptions.MissingSchema as error:
            raise TridentLoginError(
                "Aquarium URL {0} incorrectly formatted. {1}".format(
                    self.aquarium_url, error.args[0]
                )
            )
        except requests.exceptions.ConnectTimeout:
            raise TridentTimeoutError(
                "Either Aquarium took too long to respond during login or your internet"
                " connection is slow. Make sure the url {} is correct."
                " Alternatively, use Session.set_timeout"
                " to increase the request timeout.".format(self.aquarium_url)
            )
        except requests.exceptions.ConnectionError:
            raise TridentLoginError(
                "Could not contact '{}'. Login request returned no response".format(
                    self.aquarium_url
                )
            )

    @staticmethod
    def _serialize_request(url, method, body):
        return json.dumps({"url": url, "method": method, "body": body}, sort_keys=True)

    def request(self, method, path, timeout=None, allow_none=True, **kwargs):
        """
        Performs a http request.

        :param method: request method (e.g. 'put', 'post', 'get', etc.)
        :type method: str
        :param path: url to perform the request
        :type path: str
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False will raise error when json_data
                contains a None or null value (default: True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """

        url = url_build(self.aquarium_url, path)
        if not self._using_requests:
            raise ForbiddenRequestError(
                "Attempted a request ({} {}) when requests have been turned OFF.\nDATA: {}".format(
                    method.upper(), url, kwargs["json"]
                )
            )

        if timeout is None:
            timeout = self.timeout
        if not allow_none and "json" in kwargs:
            self._disallow_null_in_json(kwargs["json"])

        self.num_requests += 1
        response = requests.request(
            method, url, timeout=timeout, cookies=self.cookies, **kwargs
        )

        self.log.info(self._format_response_info(response))
        if response.status_code >= 400:
            response_info = self._format_response_info(response)
            request_status = self._format_request_status(response)
            msg = "\n".join(
                [
                    "The Aquarium server returned an error.",
                    request_status,
                    response_info,
                ]
            )
            raise TridentRequestError(msg, response)

        return self._response_to_json(response)

    def _response_to_json(self, response):
        """
        Turns :class:`requests.Request` instance into a json.
        Raises TridentRequestError if an error occurs.
        """

        if response.url == url_build(self.aquarium_url, "signin"):
            msg = (
                "There was an error with authenticating the request. Aquarium "
                + "re-routed to the sign-in page."
            )
            raise TridentRequestError(msg, response)

        try:
            response_json = response.json()
        except json.JSONDecodeError:
            msg = "Response is not JSON formatted"
            msg += "\nMessage:\n" + response.text
            self.log.error(self._format_response_info(response))
            raise TridentRequestError(msg, response)
        if response_json:
            if "errors" in response_json:
                errors = response_json["errors"]
                if isinstance(errors, list):
                    errors = "\n".join(errors)
                msg = "Error response:\n{}".format(errors)
                raise TridentRequestError(msg, response)
        return response_json

    @staticmethod
    def _disallow_null_in_json(json_data):
        """
        Raises :class:pydent.exceptions.TridentJSONDataIncomplete exception if
        json data being sent contains a null value
        """
        if None in json_data.values():
            raise TridentJSONDataIncomplete(
                "JSON data {} contains a null value.".format(json_data)
            )

    def post(self, path, json_data=None, timeout=None, allow_none=True, **kwargs):
        """
        Make a post request to the session

        :param path: url
        :type path: str
        :param json_data: json_data to post
        :type json_data: dict
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False throw error if json_data contains a null
                or None value (default True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request(
            "post",
            path,
            json=json_data,
            timeout=timeout,
            allow_none=allow_none,
            **kwargs
        )

    def put(self, path, json_data=None, timeout=None, allow_none=True, **kwargs):
        """
        Make a put request to the session

        :param path: url
        :type path: str
        :param json_data: json_data to post
        :type json_data: dict
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False throw error when json_data contains a null
                or None value (default True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request(
            "put",
            path,
            json=json_data,
            timeout=timeout,
            allow_none=allow_none,
            **kwargs
        )

    def get(self, path, timeout=None, allow_none=True, **kwargs):
        """
        Make a get request to the session

        :param path: url
        :type path: str
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False throw error when json_data contains a null
                or None value (default: True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request(
            "get", path, timeout=timeout, allow_none=allow_none, **kwargs
        )

    def delete(self, path, timeout=None, **kwargs):
        return self.request("delete", path, timeout=timeout, **kwargs)

    def __repr__(self):
        return "<{}(user='{}', url='{}')>".format(
            self.__class__.__name__, self.login, self.aquarium_url
        )

    def __str__(self):
        return self.__repr__()



"""
Models related to field values and sample properties.
"""

import json

from pydent.base import ModelBase
from pydent.exceptions import AquariumModelError
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import JSONSaveMixin, JSONDeleteMixin
from pydent.models.field_value_mixins import FieldMixin
from pydent.models.inventory import Item, Collection, ObjectType
from pydent.models.sample import Sample
from pydent.relationships import HasOne, Function, Raw, fields, HasMany
from pydent.utils import filter_list


@add_schema
class AllowableFieldType(ModelBase):
    """A AllowableFieldType model"""

    fields = dict(
        field_type=HasOne("FieldType"),
        object_type=HasOne("ObjectType"),
        sample_type=HasOne("SampleType"),
    )

    def __init__(self, field_type=None, object_type=None, sample_type=None):
        super().__init__(
            field_type_id=None,
            sample_type_id=None,
            object_type_id=None,
            field_type=field_type,
            object_type=object_type,
            sample_type=sample_type,
        )

    def __str__(self):
        return self._to_str("sample_type", "object_type")


@add_schema
class FieldType(FieldMixin, ModelBase):
    """A FieldType model"""

    fields = dict(
        allowable_field_types=HasMany("AllowableFieldType", "FieldType"),
        operation_type=HasOne(
            "OperationType", callback="find_field_parent", ref="parent_id"
        ),
        sample_type=HasOne("SampleType", callback="find_field_parent", ref="parent_id"),
        field_values=HasMany("FieldValue", "FieldType"),
    )

    def __init__(
        self,
        name=None,
        ftype=None,
        array=None,
        choices=None,
        operation_type=None,
        preferred_field_type_id=None,
        preferred_operation_type_id=None,
        required=None,
        routing=None,
        role=None,
        parent_class=None,
        parent_id=None,
        sample_type=None,
        aft_stype_and_objtype=(),
        allowable_field_types=None,
    ):
        if operation_type and sample_type:
            raise Exception(
                "Cannot instantiate a FieldType for both a OperationType and SampleType."
            )
        if operation_type:
            parent_class = "OperationType"
        if sample_type:
            parent_class = "SampleType"
        super().__init__(
            name=name,
            ftype=ftype,
            array=array,
            choices=choices,
            role=role,
            preferred_field_type_id=preferred_field_type_id,
            preferred_operation_type_id=preferred_operation_type_id,
            required=required,
            routing=routing,
            parent_class=parent_class,
            parent_id=parent_id,
            sample_type=sample_type,
            operation_type=operation_type,
            allowable_field_types=allowable_field_types,
        )
        self.part = None
        if allowable_field_types is None:
            if aft_stype_and_objtype is not None:
                for smple_type, obj_type in aft_stype_and_objtype:
                    self.create_allowable_field_type(smple_type, obj_type)

    def get_choices(self):
        if self.choices == "":
            return None
        if self.choices is not None:
            return self.choices.split(",")

    def is_parameter(self):
        return self.ftype != "sample"

    def get_allowable_field_types(self, sample_type_id=None, object_type_id=None):
        afts = self.allowable_field_types
        if afts is None or afts == []:
            self.allowable_field_types = None
            afts = self.allowable_field_types
        return filter_list(
            afts, sample_type_id=sample_type_id, object_type_id=object_type_id
        )

    def create_allowable_field_type(self, sample_type=None, object_type=None):
        afts = []
        if self.allowable_field_types:
            afts = self.allowable_field_types
        field_type = self.session.AllowableFieldType.new(
            field_type=self, sample_type=sample_type, object_type=object_type
        )
        afts.append(field_type)
        self.allowable_field_types = afts
        return field_type

    def initialize_field_value(self, field_value=None, parent=None):
        """
        Updates or initializes a new :class:`FieldValue` from this FieldType

        :param field_value: optional FieldValue to update with this FieldType
        :type field_value: FieldValue
        :return: updated FieldValue
        :rtype: FieldValue
        """

        if not field_value:
            field_value = self.session.FieldValue.new(
                name=self.name, role=self.role, field_type=self
            )
        if self.allowable_field_types:
            field_value.allowable_field_type_id = self.allowable_field_types[0].id
            field_value.allowable_field_type = self.allowable_field_types[0]
        if parent:
            field_value.set_parent(parent)
        return field_value


@add_schema
class FieldValue(FieldMixin, JSONSaveMixin, JSONDeleteMixin, ModelBase):
    """
    A FieldValue model. One of the more complex models.

    Change Log:
    * 2019_06_03 FieldValues no longer have 'wires_as_source' or 'wires_as_dest' fields. Wires may only be accessed
    via plans only or via the FieldValue instance method 'get_wires,' which accesses the FieldValues operation
    and its Plan to obtain wires.
    """

    fields = dict(
        # FieldValue relationships
        field_type=HasOne("FieldType"),
        allowable_field_type=HasOne("AllowableFieldType"),
        array=fields.Field(),
        # array=Function(lambda fv: fv.array, callback_args=(fields.Callback.SELF,)),
        item=HasOne("Item", ref="child_item_id"),
        sample=HasOne("Sample", ref="child_sample_id"),
        object_type_id=Raw(),
        object_type=HasOne("ObjectType"),
        operation=HasOne("Operation", callback="find_field_parent", ref="parent_id"),
        parent_sample=HasOne("Sample", callback="find_field_parent", ref="parent_id"),
        sid=Function("get_sid"),
        child_sample_name=Function(
            lambda fv: fv.sid, callback_args=(fields.Callback.SELF,)
        ),
        wires_as_source=HasMany("Wire", ref="from_id"),
        wires_as_dest=HasMany("Wire", ref="to_id"),
        # allowable_child_types=Function('get_allowable_child_types'),
    )

    def __init__(
        self,
        name=None,
        role=None,
        parent_class=None,
        parent_id=None,
        field_type=None,
        sample=None,
        value=None,
        item=None,
        container=None,
    ):
        """

        :param value:
        :type value:
        :param sample:
        :type sample:
        :param container:
        :type container:
        :param item:
        :type item:
        """
        super().__init__(
            name=name,
            role=role,
            parent_class=parent_class,
            parent_id=parent_id,
            field_type_id=None,
            field_type=field_type,
            child_sample_id=None,
            sample=sample,
            value=value,
            child_item_id=None,
            item=item,
            object_type=container,
            allowable_field_type_id=None,
            allowable_field_type=None,
            column=None,
            row=None,
        )

        if field_type is not None:
            self.set_field_type(field_type)

        if any([value, sample, item, container]):
            self._set_helper(
                value=value, sample=sample, item=item, object_type=container
            )

        if self.parent_class == "Operation" and not self.role:
            raise AquariumModelError(
                "FieldValue {} needs a role to be a field value for an operation".format(
                    self
                )
            )

    def get_sid(self):
        """The FieldValues sample identifier."""
        if self.sample is not None:
            return self.sample.identifier

    def get_wires(self):
        if not self.role or self.parent_class != "Operation":
            return None
        elif self.operation and self.operation.plan:
            wires = self.operation.plan.wires
            if self.role == "input":
                return [w for w in wires if w.does_wire_to(self)]
            elif self.role == "output":
                return [w for w in wires if w.does_wire_from(self)]
        return []

    def incoming_wires(self):
        if self.role == "input":
            return self.get_wires()
        return []

    def outgoing_wires(self):
        if self.role == "output":
            return self.get_wires()
        return []

    @property
    def successors(self):
        if self.outgoing_wires:
            return [x.destination for x in self.outgoing_wires]
        return []

    @property
    def predecessors(self):
        if self.incoming_wires:
            return [x.source for x in self.incoming_wires]
        return []

    def show(self, pre=""):
        if self.sample:
            if self.child_item_id:
                item = " item: {}".format(self.child_item_id) + " in {}".format(
                    self.item.object_type.name
                )
            else:
                item = ""
            print(
                "{}{}.{}:{}{}".format(pre, self.role, self.name, self.sample.name, item)
            )
        elif self.value:
            print("{}{}.{}:{}".format(pre, self.role, self.name, self.value))

    def reset(self):
        """Resets the inputs of the field_value"""
        self.value = None
        self.allowable_field_type_id = None
        self.allowable_field_type = None
        self.child_item_id = None
        self.item = None
        self.child_sample_id = None
        self.sample = None
        self.row = None
        self.column = None

    def _validate_types(self, name, x, expected_types):
        if x is not None and not any([issubclass(type(x), e) for e in expected_types]):
            raise AquariumModelError(
                "Cannot set FieldValue.{} with a {}. Expected types {}".format(
                    name, type(x), expected_types
                )
            )

    def _validate_value(self, value):
        if value is not None:
            choices = self.field_type.get_choices()
            if choices is not None:
                if value not in choices and str(value) not in choices:
                    raise AquariumModelError(
                        "Value '{}' not in list of field "
                        "type choices '{}'".format(value, choices)
                    )
        try:
            json.dumps(value)
        except TypeError as e:
            raise AquariumModelError(
                "Cannot set value '{}'. "
                "Value is not json serializable.".format(value)
            ) from e

    def _validate_item_and_container(self, item, container):
        self._validate_types("item", item, [Item, Collection])
        self._validate_types("container", container, [ObjectType])
        if item and container and item.object_type_id != container.id:
            raise AquariumModelError(
                "Item {} is not in container {}".format(item.id, str(container))
            )

    def _validate_sample(self, sample):
        self._validate_types("sample", sample, [Sample])
        if sample is not None and not issubclass(type(sample), Sample):
            raise AquariumModelError(
                "Cannot set sample with a type '{}'".format(type(sample))
            )

    def _validate_sample_and_item(self, sample, item):
        if sample and hasattr(sample, "id") and item:
            if not item.sample_id == sample.id:
                raise AquariumModelError(
                    "Cannot set FieldValue {}. Item {} is not a member of sample {}".format(
                        self, item, sample
                    )
                )

    def _validate_row_and_column(self, row: int, col: int):
        if row and col and (not isinstance(row, int) or not isinstance(col, int)):
            raise AquariumModelError("Row and column must be integers.")
        if row is not None:
            if not self.field_type.is_part():
                raise AquariumModelError(
                    "Cannot set row of a non-part for {} {}".format(
                        self.role, self.name
                    )
                )
            self.row = row
        if col is not None:
            if not self.field_type.is_part():
                raise AquariumModelError(
                    "Cannot set column of a non-part for {} {}".format(
                        self.role, self.name
                    )
                )
            self.column = col

    def _validate(self, value, sample, item, object_type, row, column):
        self._validate_row_and_column(row, column)
        self._validate_value(value)
        self._validate_sample(sample)
        self._validate_item_and_container(item, object_type)
        # if not self.field_type.part:
        #     self._validate_sample_and_item(sample, item)

    def _set_helper(
        self,
        value=None,
        sample=None,
        item=None,
        object_type=None,
        row=None,
        column=None,
    ):
        self._validate(value, sample, item, object_type, row, column)
        if row is not None:
            self.row = row
        if column is not None:
            self.column = column

        if value is not None:
            self.value = value

        if item is not None:
            self.item = item
            if hasattr(item, "id"):
                self.child_item_id = item.id
            if not sample:
                sample = item.sample
        if sample is not None:
            self.sample = sample
            if hasattr(sample, "id"):
                self.child_sample_id = sample.id
        if object_type is not None:
            self.object_type = object_type

    def valid_afts(self):
        afts = self.field_type.allowable_field_types
        if self.sample is not None:
            afts = filter_list(afts, sample_type_id=self.sample.sample_type_id)
        if self.object_type is not None:
            afts = filter_list(afts, object_type_id=self.object_type.id)
        if self.item is not None and self.item.object_type_id is not None:
            afts = filter_list(afts, object_type_id=self.item.object_type_id)
        return afts

    def _raise_aft_error(self):
        aft_list = []
        for aft in self.field_type.allowable_field_types:
            st = "none"
            ot = "none"
            if aft.object_type is not None:
                ot = aft.object_type.name
            if aft.sample_type is not None:
                st = aft.sample_type.name
            aft_list.append("{}:{}".format(st, ot))
        sid = "none"
        if self.sample is not None:
            sid = self.sample.sample_type.name
        oid = "none"
        if self.object_type is not None:
            oid = self.object_type.name
        msg = "No allowable field types found for {} {} using {} {}."
        msg += " Available afts: {}"
        raise AquariumModelError(
            msg.format(self.role, self.name, sid, oid, ", ".join(aft_list))
        )

    def set_aft(self):
        afts = self.valid_afts()
        if len(afts) >= 1:
            self.set_allowable_field_type(afts[0])
        else:
            self._raise_aft_error()

    def set_value(
        self,
        value=None,
        sample=None,
        item=None,
        object_type=None,
        row=None,
        column=None,
        container=None,
    ):
        # to maintain old API
        # if container:
        #     raise DeprecationWarning("Typed parameter 'container' is now depreciated. Please use 'object_type'")
        if object_type is None and container:
            object_type = container
        self._validate(value, sample, item, container, row, column)
        self._set_helper(
            value=value,
            sample=sample,
            item=item,
            object_type=object_type,
            row=row,
            column=column,
        )

        """Sets the value of a """
        if any([sample, object_type, item]):
            self.set_aft()
        return self

    def get_parent_key(self):
        if self.parent_class == "Operation":
            return "operation"
        elif self.parent_class == "Sample":
            return "parent_sample"
        else:
            raise AquariumModelError(
                "Parent class '{}' not recognized as a FieldValue parent.".format(
                    self.parent_class
                )
            )

    def get_parent(self):
        key = self.get_parent_key()
        return getattr(self, key)

    def set_parent(self, model):
        self.parent_id = model.id
        self.parent_class = model.__class__.__name__
        setattr(self, self.get_parent_key(), model)
        return self

    def set_operation(self, operation):
        self.parent_class = "Operation"
        if hasattr(operation, "id"):
            self.parent_id = operation.id
        self.operation = operation

    def set_sample(self, sample):
        self.parent_class = "Sample"
        if hasattr(sample, "id"):
            self.parent_id = sample.id
        self.sample = sample

    def set_field_type(self, field_type):
        """
        Sets properties from a field_type
        """

        self.field_type = field_type
        self.field_type_id = field_type.id

        if field_type.name:
            self.name = field_type.name
        if field_type.parent_class:
            if field_type.parent_class == "OperationType":
                self.parent_class = "Operation"
            elif field_type.parent_class == "SampleType":
                self.parent_class = "Sample"
            else:
                raise AquariumModelError(
                    "FieldType parent_class '{}' not recognized".format(
                        field_type.parent_class
                    )
                )
        if field_type.role:
            self.role = field_type.role

    def set_allowable_field_type(self, allowable_field_type):
        self.allowable_field_type = allowable_field_type
        self.allowable_field_type_id = allowable_field_type.id

    def choose_item(self, first=True):
        """Set the item associated with the field value"""
        index = 0
        if not first:
            index = -1
        items = self.compatible_items()
        if items is not None and len(items) > 0:
            item = items[index]
            self.set_value(item=item)
            return item
        return None

    def compatible_items(self):
        """Find items compatible with the field value"""
        return self.session.utils.compatible_items(
            self.sample.id, self.allowable_field_type.object_type_id
        )

    def __str__(self):
        return self._to_str("id", "name", "role")



"""
Model serialization/deserialization schema
"""

import inspect

from pydent.marshaller.descriptors import DataAccessor
from pydent.marshaller.exceptions import (
    MultipleValidationError,
    CallbackValidationError,
)
from pydent.marshaller.fields import Field, Callback
from pydent.marshaller.registry import SchemaRegistry
from pydent.marshaller.utils import make_signature_str


class DynamicSchema(metaclass=SchemaRegistry):
    """
    A dynamically added schema. Should be added using '@add_schema` decorator.
    """

    ignore = ()
    fields = dict()

    @classmethod
    def _get_model_fields(cls):
        return cls.fields

    @classmethod
    def init_field_accessors(cls):
        """Initializes callback accessors.

        :return:
        :rtype:
        """
        for field_name, field in cls._get_model_fields().items():
            field.register(field_name, cls.model_class)

    @classmethod
    def init_data_accessors(cls, instance, data, add_extra=True):
        """Initializes data accessors.

        :param instance:
        :type instance:
        :param data:
        :type data:
        :return:
        :rtype:
        """
        if cls.model_class is not instance.__class__:
            raise Exception("Instance and model class are different")
        for k, v in dict(data).items():
            if k in cls.ignore:
                data.pop(k, None)
                continue
            if not add_extra and k not in cls.model_class.fields:
                raise AttributeError(
                    "Expected field missing. Cannot initialize accessor for '{}' for '{}' because it is not a field."
                    " It may be missing from the 'field' dictionary.".format(
                        k, cls.model_class
                    )
                )
            if k not in cls.model_class.__dict__:
                setattr(cls.model_class, k, DataAccessor(k, cls.model_class._data_key))
            try:
                setattr(instance, k, v)
            except AttributeError as e:
                raise e
                # raise AttributeError("can't set attribute '{key}' (accessor '{objtype}.{key}' is a {accessor} type) for '{objtype}'"
                #                      " because:\n{e}"
                # .format(
                #     key=k,
                #     objtype=instance.__class__,
                #     accessor=instance.__class__.__dict__.get(k, None),
                #     e="{}: {}".format(e.__class__.__name__, e)
                # )) from e

    @classmethod
    def validate_callbacks(cls):
        """Validates expected callback signature found in any callback in the
        registered model fields.

        :return: None
        :rtype: None
        :raises: ModelValidationError
        """
        missing_callbacks = []
        not_callable = []
        wrong_signature = []
        model_class = cls.model_class
        for rname, callback_field in cls.grouped_fields[Callback.__name__].items():
            callback_func = callback_field.callback
            if callback_func is None:
                missing_callbacks.append(
                    "Callback for {model}.{name} cannot be None".format(
                        model=model_class.__name__, name=rname
                    )
                )
                continue
            error_prefix = "Callback '{callback}' for relationship '{model}.{name}'".format(
                callback=callback_func, model=model_class.__name__, name=rname
            )
            if not callable(callback_func) and not hasattr(model_class, callback_func):
                if callback_func not in missing_callbacks:
                    missing_callbacks.append(error_prefix + " is missing.")
            else:
                args_to_send = list(callback_field.callback_args)
                kwargs_to_send = dict(callback_field.callback_kwargs)
                if isinstance(callback_func, str):
                    callback_func = getattr(model_class, callback_func)
                    args_to_send = ["self"] + args_to_send
                if not callable(callback_func):
                    not_callable.append(error_prefix + " is not callable.")
                else:
                    signature = inspect.getfullargspec(callback_func)
                    expected_args = signature.args
                    expected_kwargs = {}
                    if signature.defaults:
                        default_args = expected_args[-len(signature.defaults) :]
                        expected_args = expected_args[: -len(signature.defaults)]
                        expected_kwargs = dict(zip(default_args, signature.defaults))

                    if (
                        len(expected_args) != len(args_to_send)
                        and not signature.varargs
                    ):
                        wrong_signature.append(
                            error_prefix
                            + " expects arguments {receive_args} but would receive arguments {sent_args}.".format(
                                receive_args=make_signature_str(
                                    expected_args, expected_kwargs
                                ),
                                sent_args=make_signature_str(
                                    list(args_to_send), kwargs_to_send
                                ),
                            )
                        )
                    else:
                        for k in kwargs_to_send:
                            invalid_keys = []
                            if k not in expected_kwargs and not signature.varkw:
                                invalid_keys.append(k)
                            if invalid_keys:
                                wrong_signature.append(
                                    error_prefix
                                    + " does not recognize named key(s) {invalid_keys} from signature {receive_args}.".format(
                                        invalid_keys=", ".join(
                                            ['"{}"'.format(key) for key in invalid_keys]
                                        ),
                                        receive_args=make_signature_str(
                                            expected_args, expected_kwargs
                                        ),
                                    )
                                )
                                wrong_signature.append(callback_func)

        with MultipleValidationError("") as e:
            if missing_callbacks:
                for w in missing_callbacks:
                    e.r(CallbackValidationError(w))
            if not_callable:
                for w in not_callable:
                    e.r(CallbackValidationError(w))
            if wrong_signature:
                for w in wrong_signature:
                    e.r(CallbackValidationError(w))

    @classmethod
    def register(cls, model_class):
        """Registers the schema to a model class. Saves the schema class to the
        class attribute.

        :param model_class: a model class
        :type model_class: type
        :return: None
        :rtype: None
        """
        setattr(model_class, "_model_schema", cls)

        schema_fields = {}
        ignored_fields = {}
        fields_key = model_class._fields_key
        if hasattr(model_class, fields_key):
            model_fields = getattr(model_class, fields_key)

            ignore = model_fields.pop("ignore", ())
            if isinstance(ignore, str):
                ignore = (ignore,)
            cls.ignore = ignore

            for field_name, field in model_fields.items():
                if field_name in cls.ignore:
                    ignored_fields[field_name] = field
                elif issubclass(type(field), Field):
                    if field.data_key is None:
                        field.data_key = field_name
                    if issubclass(type(field), Field):
                        schema_fields[field_name] = field
            setattr(model_class, fields_key, schema_fields)
            setattr(model_class, "_ignored_fields", ignored_fields)
        setattr(cls, "_fields", schema_fields)
        setattr(cls, "_ignored_fields", ignored_fields)
        cls.init_field_accessors()
        cls.validate_callbacks()



"""
Session class for interacting with Trident and Aquarium

This module defines the AqSession class, which is the main entry point for
accessing the request methods.
Multiple session instances can be created using a user's Aquarium login
information.
From there, the session instance is used to make http requests
(like finding models, updating plans, etc.)

AqSession creates SessionInterfaces which can make http requests to Aquarium
but does not itself make the requests.

Interfaces are accessed by:

(1) session.<ModelName> - access the model interface
(2) session.create - access the create interface
(3) session.update - access the update interface
"""

import inspect
import timeit
import webbrowser
from decimal import Decimal

from requests.exceptions import ReadTimeout

from pydent.aqhttp import AqHTTP
from pydent.base import ModelRegistry
from pydent.interfaces import (
    QueryInterfaceABC,
    QueryInterface,
    UtilityInterface,
    BrowserInterface,
)
from pydent.models import __all__ as allmodels
from pydent.browser import Browser
from pydent.sessionabc import SessionABC
from copy import copy


class AqSession(SessionABC):
    """
    Holds an AqHTTP with login information.
    Creates SessionInterfaces for models.

    .. code-block:: python

    session1 = AqSession(username, password, aquairum_url)
    session1.User.find(1)
    # <User(id=1,...)>

    """

    def __init__(self, login, password, aquarium_url, name=None, aqhttp=None):
        """

        :param login: the Aquarium login for the user
        :type login: str
        :param password: the password for the Aquarium login. This will not be
            stored anywhere. However, login headers for Aquarium will be stored
            in .__aqhttp. This may change in the future with changes
            authentication for the Aquarium API
        :type password: str
        :param aquarium_url: the http formatted Aquarium url
        :type aquarium_url: str
        :param name: (optional) name for this session
        :type name: str or None
        """
        self.name = name
        self._aqhttp = None
        if login is None and password is None and aquarium_url is None:
            if aqhttp is not None:
                self._aqhttp = aqhttp
            else:
                raise ValueError(
                    "Need either a name, password, and url OR an aqhttp instance."
                )
        else:
            self._aqhttp = AqHTTP(login, password, aquarium_url)
        self._current_user = None
        self._interface_class = QueryInterface
        self.initialize_interfaces()
        self._browser = None
        self._using_cache = False
        self.init_cache()

    @property
    def interface_class(self):
        return self._interface_class

    @interface_class.setter
    def interface_class(self, c):
        if not issubclass(c, QueryInterfaceABC):
            raise ValueError(
                "Interface {} is not a subclass of {}".format(
                    type(c), QueryInterfaceABC
                )
            )
        self._interface_class = c

    def initialize_interfaces(self):
        # initialize model interfaces
        for model_name in allmodels:
            self._register_interface(model_name)

    def open(self):
        webbrowser.open(self._aqhttp.url)

    @property
    def session(self):
        return self

    def set_verbose(self, verbose, tb_limit=None):
        self._aqhttp.log.set_verbose(verbose, tb_limit=tb_limit)

    def _log_to_aqhttp(self, msg):
        """Sends a log message to the aqhttp's logger"""
        self._aqhttp.log.info(msg)

    def _register_interface(self, model_name):
        # get model interface from model class
        model_interface = self.interface_class(model_name, self._aqhttp, self)

        # set interface to session attribute (e.g. session.Sample calls Sample model interface)
        setattr(self, model_name, model_interface)

    def set_timeout(self, timeout_in_seconds):
        """Sets the request timeout."""
        self._aqhttp.timeout = timeout_in_seconds

    @property
    def url(self):
        """Returns the aquarium_url for this session"""
        return self._aqhttp.aquarium_url

    @property
    def login(self):
        """
        Logs into aquarium, generating the necessary headers to perform requests
        to Aquarium
        """
        return self._aqhttp.login

    @property
    def current_user(self):
        """
        Returns the current User associated with this session. Returns None
        if no user is found (as in cases where the Aquarium connection is down).
        """
        if self._current_user is None:
            user = self.User.where({"login": self._aqhttp.login})
            if not user:
                return None
            self._current_user = self.User.where({"login": self._aqhttp.login})[0]
        return self._current_user

    def logged_in(self):
        """
        Returns whether the user is logged in. If the session
        is able to return the User model instance using the
        session's login credentials, the user is considered
        to be logged in.

        :return: whether user is currently logged in
        :rtype: boolean
        """
        if self.current_user is None:
            return False
        return True

    @property
    def models(self):
        """Returns list of all models available"""
        return list(ModelRegistry.models.keys())

    def model_interface(self, model_name, interface_class=None):
        """Returns model interface by name"""
        if interface_class is None:
            interface_class = self.interface_class
        return interface_class(model_name, self._aqhttp, self)

    @property
    def utils(self):
        """Instantiates a utility interface"""
        return UtilityInterface(self._aqhttp, self)

    # TODO: put 'ping' in documentation
    def ping(self, num=5):
        """Performs a number of simple requests (pings) to estimate the speed of the server.
        Displays a message about the average time each ping took."""
        try:
            ping_function = lambda: self.User.find(1)
            ping_function_source = inspect.getsource(ping_function).strip()
            secs = timeit.timeit(ping_function, number=num)
            print(
                "{} pings (using the function '{}')".format(num, ping_function_source)
            )
            print("{} seconds per ping".format("%.2E" % Decimal(secs / num)))
            return secs
        except ReadTimeout as e:
            print("Error: {}".format(e))
            print(
                "Aquarium ({}) looks like its down. The function '{}' raised a {} exception, but should not have.".format(
                    self.url, ping_function_source, ReadTimeout
                )
            )
            return None

    @property
    def browser(self):
        return self._browser

    def init_cache(self):
        self._browser = Browser(self)

    def clear_cache(self):
        self.browser.clear()

    @property
    def using_requests(self):
        return self._aqhttp._using_requests

    @using_requests.setter
    def using_requests(self, b):
        if b:
            self._aqhttp.on()
        else:
            self._aqhttp.off()

    @property
    def using_cache(self):
        return self._using_cache

    @using_cache.setter
    def using_cache(self, b):
        if b:
            self.interface_class = BrowserInterface
            if self.browser is None:
                self.init_cache()
            self.initialize_interfaces()
            self._using_cache = True
        else:
            self._using_cache = False
            self.interface_class = QueryInterface
            self.initialize_interfaces()

    def copy(self):
        instance = self.__class__(
            None, None, None, self.name, aqhttp=copy(self._aqhttp)
        )
        instance.using_requests = self.using_requests
        instance.using_cache = self.using_cache
        return instance

    def with_cache(
        self, using_requests=None, using_models=False, timeout=None, verbose=None
    ):
        return self(
            using_cache=True,
            using_models=using_models,
            using_requests=using_requests,
            timeout=timeout,
            using_verbose=verbose,
        )

    def with_requests_off(
        self, using_cache=None, using_models=True, timeout=None, verbose=None
    ):
        return self(
            using_cache=using_cache,
            using_models=using_models,
            using_requests=False,
            timeout=timeout,
            using_verbose=verbose,
        )

    @staticmethod
    def _swap_sessions(from_session, to_session):
        """Moves models from one sesssion to another."""
        models = from_session.browser.models
        if to_session.browser:
            to_session.browser.update_cache(models)
        for m in models:
            m._session = to_session

    def __call__(
        self,
        using_cache=None,
        using_requests=None,
        timeout=None,
        using_models=None,
        using_verbose=None,
    ):
        new_session = self.copy()
        new_session.parent_session = self
        if using_cache is not None:
            new_session.using_cache = using_cache
        if using_requests is not None:
            new_session.using_requests = using_requests
        if timeout is not None:
            new_session.set_timeout(timeout)
        if using_models and self.browser:
            new_session.browser.update_cache(self.browser.models)
        if using_verbose is not None:
            new_session.set_verbose(using_verbose)
        return new_session

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self, "parent_session"):
            self._swap_sessions(self, self.parent_session)

    def __repr__(self):
        return "<{}(name={}, AqHTTP={}))>".format(
            self.__class__.__name__, self.name, self._aqhttp
        )



"""
Trident exceptions
"""

from requests.exceptions import RequestException, ConnectTimeout


class TridentRequestError(IOError):
    """There was an ambiguous exception that occurred handling your request."""


class AquariumError(IOError):
    """Aquarium raised an error"""


class ForbiddenRequestError(Exception):
    """Raised when Trident attempts to make a request after requests have been
    explicitly turned off."""


class TridentJSONDataIncomplete(RequestException):
    """JSON data contains a null value and may be incomplete."""


class TridentLoginError(RequestException):
    """Trident is not properly connected to the server.
     Verify login credentials are correct."""


class TridentTimeoutError(ConnectTimeout):
    """Trident took too long to respond"""


class TridentModelNotFoundError(AttributeError):
    """Trident could not find model in list of models."""


class AquariumModelError(Exception):
    """An error occurred with this Aquarium model"""


class NoSessionError(Exception):
    """There was no session attached to the model, but one is required."""



"""
Descriptors
"""

from enum import Enum, auto


class Placeholders(Enum):
    DATA = auto()
    MARSHALL = auto()
    CALLBACK = auto()
    DEFAULT = auto()


class DataAccessor(object):
    """A descriptor that will dynamically access an instance's dictionary named
    by the `accessor` key. If the key is not in the dictionary or the value
    received from the dictionary is a :class:`Placeholders.DATA` enumerator, an
    AttributeError is raised.

    **Usage**:

    This may be used to assign dynamic properties to a class method as in
    the example below, or subclasses of the DataAccessor can be created
    to create 'hooks' when the descriptor is accessed, set, or delete.

    .. code-block::

        model_class.x = DataAccessor('x')
        model_class.y = DataAccessor('y')
        instance = model_class()

        instance.data # {'x': Placeholders.DATA, 'y': Placeholders.DATA}

        # accessing the default values
        try:
            instance.x
        except AttributeError:
            print("x being tracked but is not set")

        # setting and accessing a value
        instance.x = 5
        instance.data = {'x': 5, 'y': Placeholders.DATA}
        instance.x  # returns 5

        # raising AttributeError
        try:
            instance.y
        except AttributeError:
            print("y is not set")

        # setting the value
        instance.y = 10
        assert instance.data == {'x': 5, 'y': 10}
        assert instance.y == 10

        # deleting the value
        del instance.y
        assert instance.data == {'x': 5, 'y': Placeholders.DATA}
        try:
            instance.y
        except AttributeError:
            print("y is not set")
    """

    __slots__ = ["name", "accessor", "default"]
    HOLDER = Placeholders.DATA

    def __init__(self, name, accessor=None, default=HOLDER):
        self.name = name
        self.accessor = accessor
        if default is Placeholders.DEFAULT:
            default = self.HOLDER
        self.default = default
        if self.name == self.accessor:
            raise Exception(
                "Descriptor name '{}' cannot be accessor name '{}'".format(
                    self.name, self.accessor
                )
            )

    def get_val(self, obj):
        access_data = getattr(obj, self.accessor)
        return access_data.get(self.name, self.default)

    def __get__(self, obj, objtype):
        val = self.get_val(obj)
        if val is self.HOLDER:
            raise AttributeError(
                "type object '{}' does not have attribute '{}'".format(
                    obj.__class__.__name__, self.name
                )
            )
        else:
            return val

    def __set__(self, obj, val):
        getattr(obj, self.accessor)[self.name] = val

    def __delete__(self, obj):
        getattr(obj, self.accessor)[self.name] = self.HOLDER


class MarshallingAccessor(DataAccessor):
    """
    A generic Marshalling descriptor.
    """

    __slots__ = ["name", "accessor", "field", "deserialized_accessor", "default"]
    HOLDER = Placeholders.MARSHALL

    def __init__(self, name, field, accessor, deserialized_accessor, default=HOLDER):
        super().__init__(name, accessor, default=default)
        self.deserialized_accessor = deserialized_accessor
        if self.accessor == self.deserialized_accessor:
            raise Exception(
                "Descriptor accessor '{}' cannot be deserialized accessor '{}'".format(
                    self.accessor, self.deserialized_accessor
                )
            )
        self.field = field

    def get_val(self, obj):
        try:
            return getattr(obj, self.deserialized_accessor).get(self.name, self.default)
        except AttributeError as e:
            raise e
        except Exception as e:
            raise Exception(
                "Error retrieving attribute '{}' from '{}' because:\n".format(
                    self.name, obj.__class__
                )
                + "{}: ".format(e.__class__.__name__)
                + str(e)
            ) from e

    def __get__(self, obj, objtype):
        val = self.get_val(obj)
        if val is self.HOLDER:
            val = getattr(obj, self.accessor).get(self.name, self.HOLDER)
            if val is self.HOLDER:
                raise AttributeError(
                    "type object '{}' does not have attribute '{}'".format(
                        obj.__class__.__name__, self.name
                    )
                )
            else:
                return self.field.deserialize(obj, val)
        return val

    def __set__(self, obj, val):
        try:
            deserialized = self.field.deserialize(obj, val)
            serialized = self.field.serialize(obj, deserialized)
            getattr(obj, self.deserialized_accessor)[self.name] = deserialized
            getattr(obj, self.accessor)[self.name] = serialized
        except Exception as e:
            from traceback import format_tb

            print("__set__ traceback:")
            print("\n".join(format_tb(e.__traceback__)))
            raise Exception(
                "can't set attribute '{}' for '{}' to '{}' due to:\n{}. See the traceback"
                " printed above".format(self.name, obj, val, str(e))
            ) from e

    def __delete__(self, obj):
        del getattr(obj, self.accessor)[self.name]
        del getattr(obj, self.deserialized_accessor)[self.name]


class CallbackAccessor(MarshallingAccessor):
    """A descriptor that uses a registered :class:`marshaller.fields.Callback`
    to dynamically access the value of a callback by sending the instance to
    the callback field's `fullfill` method if the descriptor is not yet set.

    If the descriptor is already set, return that value. Deleting the
    descriptor sets the value to the default
    :class:`Placeholders.CALLBACK`, which will attempt to `fullfill` the
    descriptor once accessed.
    """

    __slots__ = ["name", "accessor", "field", "deserialized_accessor", "default"]
    HOLDER = Placeholders.CALLBACK

    def __init__(self, name, field, accessor, deserialized_accessor, default=HOLDER):
        super().__init__(name, field, accessor, deserialized_accessor, default=default)

    def __get__(self, obj, objtype):
        val = self.get_val(obj)
        if val is self.HOLDER:
            val = self.field.fullfill(obj)
        return val

    def __set__(self, obj, val):
        getattr(obj, self.deserialized_accessor)[self.name] = val
        getattr(obj, self.accessor)[self.name] = self.field.serialize(obj, val)


class RelationshipAccessor(CallbackAccessor):
    """
    The descriptor for a :class:`pydent.marshaller.fields.Relationship` field
    """

    def __set__(self, obj, val):
        deserialized = self.field.deserialize(obj, val)
        serialized = self.field.serialize(obj, deserialized)
        getattr(obj, self.deserialized_accessor)[self.name] = deserialized
        getattr(obj, self.accessor)[self.name] = serialized



"""

.. module:: pydent

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    aqhttp
    aqsession
    base
    exceptions
    interfaces
    models
    relationships
    browser
    planner
    marshaller
    utils

"""

# without nest_asyncio, Jupyter notebooks will crash if an async method is used
import nest_asyncio

nest_asyncio.apply()

from .aqsession import AqSession
from .base import ModelBase, ModelRegistry
from .utils import pprint
from .browser import Browser
from .__version__ import __version__, __title__, __authors__, __homepage__, __repo__
import getpass


def login(user, url, password=None):
    """
    Create a new AqSession instance from login information.

    :param user: user login
    :type user: basestring
    :param url: server url
    :type url: basestring
    :param password: optional password. If left blank, a secure prompt will ask for the password.
    :type password: basestring | None
    :return: the session
    :rtype: AqSession
    """
    if password is None:
        return AqSession(user, getpass.getpass(), url)
    return AqSession(user, password, url)



"""
Models related to inventory, like Items, Collections, ObjectTypes, and PartAssociations
"""

from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.models.crud_mixin import SaveMixin
from pydent.models.data_associations import DataAssociatorMixin
from pydent.relationships import Raw, HasOne, HasMany, HasManyThrough, HasManyGeneric


@add_schema
class Item(DataAssociatorMixin, SaveMixin, ModelBase):
    """
    A physical object in the lab, which a location and unique id.
    """

    fields = dict(
        sample=HasOne("Sample"),
        object_type=HasOne("ObjectType"),
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Item"}
        ),
        data=Raw(),
        ignore=("locator_id",),
        part_associations=HasMany(
            "PartAssociation", ref="part_id"
        ),  # TODO: add to change log
        collections=HasManyThrough(
            "Collection", "PartAssociation"
        ),  # TODO: add to change log
    )
    query_hook = {"methods": ["is_part"]}

    def __init__(
        self=None, sample_id=None, sample=None, object_type=None, object_type_id=None
    ):
        super().__init__(
            object_type_id=object_type_id,
            object_type=object_type,
            sample_id=sample_id,
            sample=sample,
        )

    def make(self):
        """Makes the Item on the Aquarium server. Requires
        this Item to be connected to a session."""
        result = self.session.utils.create_items([self])
        return self.reload(result[0]["item"])

    def save(self):
        """A synonym for `make`"""
        return self.make()

    def is_deleted(self):
        return self.location == "deleted"

    @property
    def containing_collection(self):
        """
        Returns the collection of which this Item is a part.

        Returns the collection object if the Item is a part, otherwise
        returns None.
        """
        if not self.is_part:
            return None

        assoc_list = self.session.PartAssociation.where({"part_id": self.id})
        if not assoc_list:
            return

        if len(assoc_list) != 1:
            return None

        part_assoc = next(iter(assoc_list))
        if not part_assoc:
            return None

        return self.session.Collection.find(part_assoc.collection_id)

    def as_collection(self):
        """
        Returns the Collection object with the ID of this Item, which must be a
        collection.

        Returns None if this Item is not a collection.
        """
        if not self.is_collection:
            return None

        return self.session.Collection.find(self.id)

    @property
    def is_collection(self):
        """
        Returns True if this Item is a collection in a PartAssociation.

        Note: this is not how Aquarium does this test in the `collection?` method.
        """
        assoc_list = self.session.PartAssociation.where({"collection_id": self.id})
        return bool(assoc_list)

    # TODO: add to change log
    @property
    def collection(self):
        return self.collections[0]

    # TODO: add to change log
    @property
    def part_association(self):
        return self.part_associations[0]


@add_schema
class Collection(
    DataAssociatorMixin, ModelBase
):  # pylint: disable=too-few-public-methods
    """A Collection model, such as a 96-well plate, which contains many `parts`, each
    of which can be associated with a different sample."""

    fields = dict(
        object_type=HasOne("ObjectType"),
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Collection"}
        ),
        part_associations=HasMany("PartAssociation", "Collection"),
        parts=HasManyThrough("Item", "PartAssociation", ref="part_id"),
    )
    query_hook = {"methods": ["dimensions"]}

    @property
    def matrix(self):
        """
        Returns the matrix of Samples for this Collection.

        (Consider using samples of parts directly.)
        """
        num_row, num_col = self.dimensions
        sample_matrix = list()
        for row in range(num_row):
            row_list = list()
            for col in range(num_col):
                sample_id = None
                part = self.part(row, col)
                if part:
                    sample_id = part.sample.id
                row_list.append(sample_id)
            sample_matrix.append(row_list)
        return sample_matrix

    def part(self, row, col):
        """
        Returns the part Item at (row, col) of this Collection (zero-based).
        """
        parts = [
            assoc.part
            for assoc in self.part_associations
            if assoc.row == row and assoc.column == col
        ]

        if not parts:
            return None

        return next(iter(parts))

    def as_item(self):
        """
        Returns the Item object with the ID of this Collection
        """
        return self.session.Item.find(self.id)

    def create(self):
        self.as_item().create()

    def update(self):
        self.as_item().update()

    def save(self):
        self.as_item().save()


@add_schema
class ObjectType(SaveMixin, ModelBase):
    """A ObjectType model that represents the type of container an item is."""

    fields = dict(items=HasMany("Item", "ObjectType"), sample_type=HasOne("SampleType"))

    def __str__(self):
        return self._to_str("id", "name")


@add_schema
class PartAssociation(ModelBase):
    """
    Represents a PartAssociation linking a part to a collection. Collections contain
    many `parts`, each of which can refer to a different sample.
    """

    fields = dict(part=HasOne("Item", ref="part_id"), collection=HasOne("Collection"))

    def __init__(self, part_id=None, collection_id=None, row=None, column=None):
        super().__init__(
            part_id=part_id, collection_id=collection_id, row=row, column=column
        )



"""
Serialization/Deserialization

==========
marshaller
==========

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    base
    descriptors
    exceptions
    fields
    registry
    schema

"""

from pydent.marshaller.base import SchemaModel, add_schema
from pydent.marshaller.registry import SchemaRegistry, ModelRegistry
from pydent.marshaller import fields, exceptions, descriptors



"""
PlannerLayout
"""

import networkx as nx
from collections import OrderedDict
from pydent.planner.utils import _id_getter, get_subgraphs
from pydent.utils import make_async
from pydent.models import Operation, Plan
from typing import Iterable, Tuple


class PlannerLayout(object):

    TOP_RIGHT = (100, 100)
    BOX_DELTA_X = 170
    BOX_DELTA_Y = 70
    BOX_WIDTH = 150
    BOX_HEIGHT = 70
    STATUS_COLORS = {
        "waiting": "orange",
        "error": "red",
        "pending": "yellow",
        "running": "green",
        "delayed": "magenta",
        "done": "black",
        "planning": "grey",
    }

    def __init__(self, G=None):
        if G is None:
            G = nx.DiGraph()
        self.G = G
        self.annotations = []

    @classmethod
    def from_plan(cls, plan: Plan):
        """Creates a layout from a :class:`pydent.models.Plan` instance"""
        G = nx.DiGraph()
        layout = cls(G)

        # store reference of operation
        fv_to_op_dict = {}
        if not plan.operations:
            return layout
        for op in plan.operations:
            for fv in op.field_values:
                fv_to_op_dict[fv.rid] = op

        @make_async(10, progress_bar=False)
        def add_wires(wires):
            if not wires:
                return
            missing_operations = []
            for wire in wires:
                # TODO:
                try:
                    from_id = _id_getter(fv_to_op_dict[wire.source.rid])
                    to_id = _id_getter(fv_to_op_dict[wire.destination.rid])
                except KeyError as e:
                    pass
                    raise e
                if from_id is not None and to_id is not None:
                    if from_id not in G:
                        missing_operations.append(from_id)
                    if to_id not in G:
                        missing_operations.append(to_id)
                    if from_id in G and to_id in G:
                        G.add_edge(from_id, to_id, wire=wire)
            # if missing_operations:
            #     raise Exception("The following operations are missing from the graph: {}".format(missing_operations))
            return wires

        @make_async(10, progress_bar=False)
        def add_ops(ops: Iterable[Operation]):
            if not ops:
                return
            for op in ops:
                layout._add_operation(op)
            return ops

        add_ops(plan.operations)
        add_wires(plan.wires)

        # fix operation coordinates if None
        for op in layout.operations:
            if op.x is None:
                op.x = 0
            if op.y is None:
                op.y = 0

        return layout

    @property
    def nodes(self):
        return self.G.nodes

    @property
    def operations(self):
        for node in self.G.nodes:
            if "operation" in self.G.node[node]:
                yield self.G.node[node]["operation"]

    def iter_operations(self):
        for node in self.G.nodes:
            if "operation" in self.G.node[node]:
                yield (node, self.G.node[node]["operation"])

    def _add_operation(self, operation: Operation):
        """
        Adds an Operation to the Layout

        :param op: operation
        :type op: pydent.models.Operation
        :return: None
        :rtype: None
        """
        return self.G.add_node(_id_getter(operation), operation=operation)

    def subgraph(self, nodes):
        """Returns a subgraph layout from a list of node_ids"""
        return self.__class__(G=self.G.subgraph(nodes))

    def nodes_to_ops(self, nodes):
        """Returns operations from a list of node_ids"""
        return [self.G.node[n]["operation"] for n in nodes]

    def ops_to_nodes(self, ops):
        """Returns node_ids for each operation"""
        return [_id_getter(op) for op in ops]

    def ops_to_layout(self, ops):
        """Returns a sub-layout containing only the operations"""
        return self.subgraph(self.ops_to_nodes(ops))

    def get_independent_layouts(self):
        """
        Finds all independent subgraphs.

        :return: list of PlannerLayout
        :rtype: list
        """
        subgraphs = get_subgraphs(self.G)
        sublayouts = [self.__class__(G=g) for g in subgraphs]
        return sublayouts

    def _topological_sort(self):
        subgraphs = self.get_independent_layouts()
        for subgraph in subgraphs:
            subgraph._topological_sort_helper()
        self.arrange_layouts(subgraphs)

    def topo_sort_in_place(self):
        cx, cy = self.midpoint()
        self._topological_sort()
        cx2, cy2 = self.midpoint()
        self.translate(cx - cx2, cy - cy2)

    def topo_sort(self):
        """
        Does a topological sort of the operations in this layout.

        Discovers individual subgraphs and topologically sorts each subgraph,
        and then aligns subgraphs horizontally.

        :return: None
        :rtype: None
        """
        self._topological_sort()
        self.move(*self.TOP_RIGHT)

    def prettify(self):
        return self.topo_sort()

    @classmethod
    def arrange_layouts(cls, layouts):
        x = 0
        y = 0
        for layout in layouts:
            layout.move(x, y)
            x += layout.width + cls.BOX_DELTA_X
            y = layout.y

    def to_grid(
        self, columns: int, axis: int = 1, border_x: int = None, border_y: int = None
    ):
        """
        Arrange layouts in a grid format.

        :param columns: maximum number of columns (or rows when axis=0)
        :type columns: int
        :param axis: which axis to limit. Default: 1 (columns) or 0 (rows)
        :type axis: int
        :param border_x: (optional) separation between each cell in the grid
        :type border_x: int
        :param border_y: (optional) separation between each cell in the grid
        :type border_y: int
        :return: layout
        :rtype: layout
        """
        layouts = self.get_independent_layouts()
        if len(layouts) == 1:
            self.move(100, 100)
            return self

        # make grid assignments
        assignments = []
        c = 0
        r = 0
        for layout in layouts:
            assignments.append((r, c, layout))
            c += 1
            if c >= columns:
                r += 1
                c = 0

        # find cell height and width
        if border_x is None:
            border_x = self.BOX_DELTA_X * 2
        if border_y is None:
            border_y = self.BOX_DELTA_Y * 2
        grid_cell_height = 0
        grid_cell_width = 0
        for layout in layouts:
            if layout.height > grid_cell_height:
                grid_cell_height = layout.height
            if layout.width > grid_cell_width:
                grid_cell_width = layout.width
        grid_cell_height += border_y
        grid_cell_width += border_x

        for r, c, layout in assignments:
            if axis == 0:
                r, c = c, r
            layout.center(c * grid_cell_width, r * grid_cell_height)
        self.move(100, 100)
        return self

    # TODO: minimize crossings
    def _topological_sort_helper(self):
        """Attempt a rudimentary topological sort on the plan"""

        _x, _y = self.TOP_RIGHT

        y = _y
        delta_x = self.BOX_DELTA_X
        delta_y = -self.BOX_DELTA_Y

        max_depth = {}
        roots = self.roots()
        for root in roots:
            depths = nx.single_source_shortest_path_length(self.G, root)
            for n, d in depths.items():
                max_depth[n] = max(max_depth.get(n, d), d)

        # push roots 'up' so they are not stuck on layer one
        for root in roots:
            successors = list(self.successors(root))
            if len(successors) > 0:
                min_depth = min([max_depth[s] for s in successors])
                max_depth[root] = min_depth - 1

        by_depth = OrderedDict()

        for node, depth in max_depth.items():
            by_depth.setdefault(depth, [])
            by_depth[depth].append(node)
        for depth in sorted(by_depth):
            op_ids = by_depth[depth]

            # sort by predecessor_layout
            predecessor_avg_x = []
            x = 0
            for op_id in op_ids:
                predecessors = list(self.predecessors(op_id))
                if len(predecessors) > 0:
                    x, _ = self.subgraph(predecessors).midpoint()
                    predecessor_avg_x.append((x, op_id))
                else:
                    predecessor_avg_x.append((x + 1, op_id))
            sorted_op_ids = [
                op_id for _, op_id in sorted(predecessor_avg_x, key=lambda x: x[0])
            ]

            x = _x
            # sorted_op_ids = sorted(op_ids)
            ops = self.nodes_to_ops(sorted_op_ids)
            for op in ops:
                op.x = x
                op.y = y
                x += delta_x
            layer = self.subgraph(op_ids)
            predecessor_layout = self.predecessor_layout(layer)
            layer.align_x_midpoints_to(predecessor_layout)
            y += delta_y

        # readjust
        self.move(_x, _y)

    def align_ops_with_successors(self, ops):
        """
        Aligns the Operations to their successors by midpoints.

        Used in topological sorting of plans.

        :param ops: list of :class:`pydent.models.Operation`
        :type ops: list
        :return: None
        :rtype: None
        """
        successors = self.ops_to_successors(ops)
        self.align_x_of_ops(ops, successors)

    def align_ops_with_predecessors(self, ops):
        """
        Aligns the Operations to their predecessors by midpoints.

        Used in topological sorting of plans.

        :param ops: list of :class:`pydent.models.Operation`
        :type ops: list
        :return: None
        :rtype: None
        """
        predecessors = self.ops_to_predecessors(ops)
        self.align_x_of_ops(ops, predecessors)

    def predecessors(self, node):
        """
        Return predecessor nodes

        :param node: node id
        :type node: basestring
        :return: list of node ids
        :rtype: list
        """
        return self.G.predecessors(node)

    def successors(self, node):
        """
        Return successors nodes

        :param node: node id
        :type node: basestring
        :return: list of node ids
        :rtype: list
        """
        return self.G.successors(node)

    def predecessor_layout(self, layout):
        n_bunch = self.collect_predecessors(layout.roots())
        return self.subgraph(n_bunch)

    def successor_layout(self, layout):
        n_bunch = self.collect_successors(layout.leaves())
        return self.subgraph(n_bunch)

    def collect_predecessors(self, nodes):
        """
        Return all predecessors from a list of nodes

        :param nodes: list of node ids
        :type nodes: list
        :return: list of node ids
        :rtype: list
        """
        predecessors = []
        for node in nodes:
            predecessors += self.predecessors(node)
        return list(set(predecessors))

    def collect_successors(self, nodes):
        """
        Return all successors from a list of nodes

        :param nodes: list of node ids
        :type nodes: list
        :return: list of node ids
        :rtype: list
        """
        successors = []
        for node in nodes:
            successors += self.successors(node)
        return list(set(successors))

    def ops_to_predecessors(self, ops):
        """
        Return all predecessor operations from a list of operations

        :param nodes: list of :class:`pydent.models.Operation`
        :type nodes: list
        :return: list of node ids
        :rtype: list
        """
        nodes = self.collect_predecessors(self.ops_to_nodes(ops))
        return self.nodes_to_ops(nodes)

    def ops_to_successors(self, ops):
        """
        Return all successor operations from a list of operations

        :param nodes: list of :class:`pydent.models.Operation`
        :type nodes: list
        :return: list of node ids
        :rtype: list
        """
        nodes = self.collect_successors(self.ops_to_nodes(ops))
        return self.nodes_to_ops(nodes)

    def leaves(self):
        """Returns the leaves of this layout"""
        leaves = []
        for n in self.nodes:
            if len(list(self.successors(n))) == 0:
                leaves.append(n)
        return leaves

    def roots(self):
        """Returns the roots of this layout"""
        roots = []
        for n in self.nodes:
            if len(list(self.predecessors(n))) == 0:
                roots.append(n)
        return roots

    def ops_to_leaves(self):
        return self.nodes_to_ops(self.leaves())

    def ops_to_roots(self):
        return self.nodes_to_ops(self.roots())

    def align_x_midpoints_to(self, other_layout):
        """
        Align the midpoint x-coordinate of this layout with that of another.

        :param other_layout: the other canvas layout
        :type other_layout: PlannerLayout
        :return: None
        :rtype: None
        """
        if len(other_layout.G) == 0:
            return
        other_x, _ = other_layout.midpoint()
        x, _ = self.midpoint()
        deltax = other_x - x
        self.translate(deltax, 0)

    def align_y_midpoints_to(self, other_layout):
        """
        Align this midpoint y-coordinates of this layout with another layout

        :param other_layout: the other canvas layout
        :type other_layout: PlannerLayout
        :return: None
        :rtype: None
        """
        if len(other_layout.G) == 0:
            return
        _, other_y = other_layout.midpoint()
        self.y = other_y

    def align_x_of_ops(self, ops1, ops2):
        """
        Aligns a set of Operations with another set of Operations within this
        PlannerLayout

        :param ops1: list of :class:`pydent.models.Operation`
        :type ops1: list
        :param ops2: list of :class:`pydent.models.Operation`
        :type ops2: list
        :return: None
        :rtype: None
        """
        layout1 = self.ops_to_layout(ops1)
        layout2 = self.ops_to_layout(ops2)
        return layout1.align_x_midpoints_to(layout2)

    @property
    def xy(self):
        return self.bounds()[0]

    @property
    def x(self):
        return self.xy[0]

    @property
    def y(self):
        return self.xy[1]

    @x.setter
    def x(self, new_x):
        deltax = new_x - self.x
        self.translate(deltax, 0)

    @y.setter
    def y(self, new_y):
        delta_y = new_y - self.y
        self.translate(0, delta_y)

    def bounds(self) -> Tuple[int, int]:
        """
        Returns the bounding box of the layout as the upper-left and
        lower-right coordinates.

        :return: x,y coordinate
        :rtype: tuple
        """
        x_arr = [op.x for op in self.operations]
        y_arr = [op.y for op in self.operations]
        return ((min(x_arr), min(y_arr)), (max(x_arr), max(y_arr)))

    def midpoint(self) -> Tuple[int, int]:
        """
        Returns the midpoint x,y coordinates of the layout

        :return: x,y coordinate
        :rtype: tuple
        """
        ul, lr = self.bounds()
        x, y = (lr[0] - ul[0]) / 2 + ul[0], (lr[1] - ul[1]) / 2 + ul[1]
        return x, y

    def translate(self, delta_x, delta_y):
        """
        Translates the layout by the x,y coordinates

        :param delta_x: delta x
        :type delta_x: int
        :param delta_y: delta y
        :type delta_y: int
        :return: None
        :rtype: None
        """
        for op in self.operations:
            op.x += delta_x
            op.y += delta_y
        return self.operations

    def move(self, x, y):
        self.x = x
        self.y = y

    def center(self, x, y):
        self.move(x - self.width / 2, y - self.height / 2)

    @property
    def width(self):
        return self.bounds()[1][0] - self.bounds()[0][0]

    @property
    def height(self):
        return self.bounds()[1][1] - self.bounds()[0][1]

    def pos(self):
        pos = {}
        for n, op in self.iter_operations():
            pos[n] = (op.x, -op.y)
        return pos

    def _status_colors(self):
        node_color = [
            self.STATUS_COLORS.get(op.status, "rosybrown") for op in self.operations
        ]
        return node_color

    def draw(self):
        return nx.draw(self.G, pos=self.pos(), node_color=self._status_colors())

    def __len__(self):
        return len(self.G)



import logging
import math
import pprint
import traceback
from logging import DEBUG, INFO, CRITICAL, ERROR, WARNING
from colorlog import ColoredFormatter


def new_logger(name, level=logging.ERROR):
    """Instantiate a new logger with the given name. If channel handler exists, do not create a new one."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # make stream handler
    if not logger.handlers:
        handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.tb_limit = 0
        formatter = ColoredFormatter(
            "%(log_color)s%(levelname)s - %(name)s - %(asctime)s - %(message)s",
            log_colors={
                "DEBUG": "cyan",
                "INFO": "white",
                "SUCCESS:": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    else:
        handler = logger.handlers[0]
    return logger, handler


def condense_long_lists(d, max_list_len=20):
    """
    Condense the long lists in a dictionary

    :param d: dictionary to condense
    :type d: dict
    :param max_len: max length of lists to display
    :type max_len: int
    :return:
    :rtype:
    """
    if isinstance(d, dict):
        return_dict = {}
        for k in d:
            return_dict[k] = condense_long_lists(dict(d).pop(k))
        return dict(return_dict)
    elif isinstance(d, list):
        if len(d) > max_list_len:
            g = max_list_len / 2
            return d[: math.floor(g)] + ["..."] + d[-math.ceil(g) :]
        else:
            return d[:]
    return str(d)


class Loggable(object):
    def __init__(self, inst, name=None):
        self.instance = inst
        if name is None:
            self.name = "{}(id={})".format(self.instance.__class__, id(self.instance))
        else:
            self.name = name

    @property
    def logger(self):
        return new_logger(self.name)[0]

    def set_tb_limit(self, limit):
        for h in self.logger.handlers:
            h.tb_limit = limit

    def set_log_level(self, level, tb_limit=None):
        for h in self.logger.handlers:
            h.setLevel(level)
        if tb_limit is not None:
            self.set_tb_limit(tb_limit)

    def set_verbose(self, verbose, tb_limit=0):
        if verbose:
            self.set_log_level(logging.INFO, tb_limit)
        else:
            self.set_log_level(logging.ERROR, tb_limit)

    def pprint_data(
        self, data, width=80, depth=10, max_list_len=20, compact=True, indent=1
    ):
        return pprint.pformat(
            condense_long_lists(data, max_list_len=max_list_len),
            indent=indent,
            width=width,
            depth=depth,
            compact=compact,
        )

    def log(self, msg, level):
        self.logger.log(level, msg)
        if self.logger.isEnabledFor(level):
            tb_limit = self.logger.handlers[0].tb_limit
            if tb_limit:
                traceback.print_stack(limit=tb_limit)

    def critical(self, msg):
        self.log(msg, CRITICAL)

    def error(self, msg):
        self.log(msg, ERROR)

    def warn(self, msg):
        self.log(msg, WARNING)

    def info(self, msg):
        self.log(msg, INFO)

    def debug(self, msg):
        self.log(msg, DEBUG)



"""
Fields
"""

from abc import ABC, abstractmethod
from enum import Enum, auto

from pydent.marshaller.descriptors import (
    CallbackAccessor,
    MarshallingAccessor,
    RelationshipAccessor,
    Placeholders,
)
from pydent.marshaller.exceptions import (
    AllowNoneFieldValidationError,
    RunTimeCallbackAttributeError,
)
from pydent.marshaller.registry import ModelRegistry
from pydent.marshaller.utils import make_signature_str


class FieldABC(ABC):
    """
    Field abstract base class
    """

    @abstractmethod
    def serialize(self):
        pass

    @abstractmethod
    def deserialize(self):
        pass

    @abstractmethod
    def _serialize(self):
        pass

    @abstractmethod
    def _deserialize(self):
        pass


class Field(FieldABC):
    """A serialization/deserialization field."""

    ACCESSOR = MarshallingAccessor

    def __init__(
        self, many=None, data_key=None, allow_none=None, default=Placeholders.DEFAULT
    ):
        """A standard field. Performs no functions on serialized and
        deserialized data.

        :param many: whether to treat serializations and deserializations as a per-item basis in a list
        :type many: bool
        :param data_key: the data_key, or attribute name, of this field. Calling this as an attribute to the registered
        nested should return this field's descriptor.
        :type data_key: basestring
        :param allow_none: whether to return None if None is received in deserialization or serializaiton methods
        :type allow_none: bool
        :raises AllowNoneFieldValidationError if None is received in serialize and deserialized methods
        and self.allow_none == False
        """

        if allow_none is None:
            allow_none = True
        if many is None:
            many = False
        self.many = many
        self.objtype = None
        self.data_key = data_key
        self.allow_none = allow_none
        self.default = default

    def set_data_key(self, key):
        self.data_key = key

    def _deserialize(self, owner, data):
        return data

    def _serialize(self, owner, data):
        return data

    def deserialize(self, owner, data):
        if data is None:
            if not self.allow_none:
                raise AllowNoneFieldValidationError(
                    "None is not allowed for field '{}'".format(self.data_key)
                )
            return None
        if self.many:
            for i, x in enumerate(data):
                data[i] = self._deserialize(owner, x)
            return data
        return self._deserialize(owner, data)

    def serialize(self, owner, data):
        if data is None:
            if not self.allow_none:
                raise AllowNoneFieldValidationError(
                    "None is not allowed for field '{}'".format(self.data_key)
                )
            return None
        if self.many:
            return [self._serialize(owner, d) for d in data]
        return self._serialize(owner, data)

    def register(self, name, objtype):
        """Registers the field to a nested class. Instantiates the corresponding
        descriptor (i.e. accessor)

        :param name: name of the field
        :type name: basestring
        :param objtype: the nested class to register the field to
        :type objtype: SchemaModel
        :return: None
        :rtype: None
        """

        self.objtype = objtype
        if name not in objtype.__dict__:
            key = name
            if self.data_key:
                key = self.data_key
            setattr(
                objtype,
                name,
                self.ACCESSOR(
                    name=key,
                    field=self,
                    accessor=objtype._data_key,
                    deserialized_accessor=objtype._deserialized_key,
                    default=self.default,
                ),
            )

    def __str__(self):
        return "<{cls} key='{objtype}.{key}' many={many} allow_none={allow_none}>".format(
            cls=self.__class__.__name__,
            key=self.data_key,
            many=self.many,
            allow_none=self.allow_none,
            objtype=self.objtype,
        )


class Nested(Field):
    """Represents a field that returns another nested instance."""

    def __init__(self, nested, many=None, data_key=None, allow_none=None, lazy=None):
        """Nested relationship initializer.

        :param nested: the nested name of nested field. Should exist in the ModelRegistery.
        :type nested: SchemaModel
        :param many: whether to treat serializations and deserializations as a per-item basis in a list
        :type many: bool
        :param data_key: the data_key, or attribute name, of this field. Calling this as an attribute to the registered
        nested should return this field's descriptor.
        :type data_key: basestring
        :param allow_none: whether to return None if None is received in deserialization or serializaiton methods
        :type allow_none: bool
        :param lazy: if set to True (default), perform lazy serialization and deserialization. If the data received
        by `deserialize` is the expected nested, return that data. If the object
        recieved by `serialize` is not the expected nested, return that data.
        :type lazy:
        """

        self.nested = nested
        if lazy is None:
            self.lazy = True
        if allow_none is None:
            allow_none = True
        super().__init__(many, data_key, allow_none)

    def get_model(self):
        return ModelRegistry.get_model(self.nested)

    def _deserialize(self, owner, data):
        if data is None and self.allow_none:
            return None
        elif self.lazy and isinstance(data, self.get_model()):
            return data
        return self.get_model()._set_data(data, owner)

    def _serialize(self, owner, obj):
        if obj is None and self.allow_none:
            return None
        elif self.lazy and not isinstance(obj, self.get_model()):
            return obj
        return obj._get_data()


class Callback(Field):
    """Make a callback when called."""

    class __FLAGS(Enum):
        SELF = auto()

    SELF = __FLAGS.SELF
    ACCESSOR = CallbackAccessor

    def __init__(
        self,
        callback,
        callback_args=None,
        callback_kwargs=None,
        cache=False,
        data_key=None,
        many=None,
        allow_none=None,
        always_dump=False,
    ):
        """A Callback field initializer.

        :param callback: name of the callback function or a callable. If a name, the name should exist
        as a function in the owner instance. Invalid callback signatures are captures on class creation.
        :type callback: callable|basestring
        :param callback_args: a tuple of arguments to use in the callback. If any of the callback arguments  or
        values of the callback kwargs are callable, the owner will be passed to the callable. The owner instance will
        replace any arguments that are `Callback.SElF`
        :type callback_args: tuple
        :param callback_kwargs: a dictionary of kwargs to use in the callback
        :type callback_kwargs: dict
        :param cache: whether to cache the result using `setattr` on the owner instance. This will initialize the
        serialization and deserialization procedures detailed in the corresponding field/descriptor.
        :type cache: bool
        :param many: whether to treat serializations and deserializations as a per-item basis in a list
        :type many: bool
        :param allow_none: whether to return None if None is received in deserialization or serializaiton methods
        :type allow_none: bool
        :param data_key: the data_key, or attribute name, of this field. Calling this as an attribute to the registered
        nested should return this field's descriptor.
        :param always_dump: if True, this field will be serialized by default
        :type always_dump: bool
        """

        super().__init__(many, data_key, allow_none)
        self.callback = callback
        self.always_dump = always_dump
        if callback_args is None:
            callback_args = tuple()
        elif not isinstance(callback_args, (list, tuple)):
            callback_args = (callback_args,)
        self.callback_args = tuple(callback_args)
        if callback_kwargs is None:
            callback_kwargs = {}
        self.callback_kwargs = dict(callback_kwargs)
        self.cache = cache

    def _callback_signature(self, args=None, kwargs=None):
        if args is None:
            args = self.callback_args
        if kwargs is None:
            kwargs = self.callback_kwargs
        return "{func}{args}".format(
            func=self.callback, args=make_signature_str(args, kwargs)
        )

    def get_callback_args(self, owner, extra_args=None):
        """Processes the callback args"""
        args = []
        callback_args = list(self.callback_args)
        if extra_args:
            callback_args += list(extra_args)
        try:
            for a in callback_args:
                if a is self.SELF:
                    args.append(owner)
                elif callable(a):
                    args.append(a(owner))
                else:
                    args.append(a)
        except AttributeError as e:
            raise RunTimeCallbackAttributeError(
                "There was an error retrieving callback arguments for '{sig}' due to:\n{e}".format(
                    sig=self._callback_signature(),
                    e="{}: {}".format(e.__class__.__name__, e),
                )
            ) from e
        return args

    def get_callback_kwargs(self, owner, extra_kwargs):
        """Processes the callback kwargs"""
        kwargs = {}
        callback_kwargs = dict(self.callback_kwargs)
        if extra_kwargs:
            callback_kwargs.update(extra_kwargs)
        try:
            for k, v in callback_kwargs.items():
                if callable(v):
                    kwargs[k] = v(owner)
                elif v is self.SELF:
                    kwargs[k] = owner
                else:
                    kwargs[k] = v
        except AttributeError as e:
            raise RunTimeCallbackAttributeError(
                "There was an error retrieving callback keyword arguments for '{func}(args)' due to:\n{e}".format(
                    func=self.callback,
                    args=self._callback_signature(),
                    e="{}: {}".format(e.__class__.__name__, e),
                )
            ) from e
        return kwargs

    def fullfill(self, owner, cache=None, extra_args=None, extra_kwargs=None):
        """Calls the callback function using the owner object. A
        Callback.SELF arg value will be replaced to be equivalent to the
        owner instance model.

        :param owner: the owning object
        :type owner: SchemaModel
        :return: function result
        :rtype: any
        """

        if callable(self.callback):
            func = self.callback
        else:
            func = getattr(owner, self.callback)

        callback_args = self.get_callback_args(owner, extra_args=extra_args)
        callback_kwargs = self.get_callback_kwargs(owner, extra_kwargs=extra_kwargs)

        try:
            val = func(*tuple(callback_args), **callback_kwargs)
        except AttributeError as e:
            raise RunTimeCallbackAttributeError(
                "There was an calling '{signature}' due to:\n{e}".format(
                    signature=self._callback_signature(callback_args, callback_kwargs),
                    e="{}: {}".format(e.__class__.__name__, e),
                )
            ) from e
        if cache is None:
            cache = self.cache
        if cache:
            self.cache_result(owner, val)
        return val

    def cache_result(self, owner, val):
        setattr(owner, self.data_key, val)

    def _deserialize(self, owner, data):
        raise NotImplementedError(
            "_deserialize is not implemented for field {}".format(self)
        )

    def deserialize(self, *args, **kwargs):
        raise NotImplementedError(
            "deserialize is not implemented for field {}".format(self)
        )


class Relationship(Callback):
    """
    A composition (Callback/Nested) field that uses a callback to retrieve a model.
    """

    ACCESSOR = RelationshipAccessor

    def __init__(
        self,
        nested,
        callback,
        callback_args=None,
        callback_kwargs=None,
        cache=True,
        data_key=None,
        many=None,
        allow_none=None,
        always_dump=False,
    ):
        """
        Relationship initializer

        :param nested: the nested name of nested field. Should exist in the ModelRegistery.
        :type nested: SchemaModel
        :param callback: name of the callback function or a callable. If a name, the name should exist
        as a function in the owner instance. Invalid callback signatures are captures on class creation.
        :type callback: callable|basestring
        :param callback_args: a tuple of arguments to use in the callback. If any of the callback arguments  or
                                values of the callback kwargs are callable, the owner will be passed to the callable. The owner instance will
                                replace any arguments that are `Callback.SElF`
        :type callback_args: tuple
        :param callback_kwargs: a dictionary of kwargs to use in the callback
        :type callback_kwargs: dict
        :param cache: whether to cache the result using `setattr` on the owner instance. This will initialize the
        serialization and deserialization procedures detailed in the corresponding field/descriptor.
        :type cache: bool
        :param many: whether to treat serializations and deserializations as a per-item basis in a list
        :type many: bool
        :param data_key: the data_key, or attribute name, of this field. Calling this as an attribute to the registered
        nested should return this field's descriptor.
        """

        super().__init__(
            callback,
            callback_args,
            callback_kwargs,
            cache,
            data_key,
            many,
            allow_none,
            always_dump,
        )
        self.nested_field = Nested(nested, many, data_key)
        self.nested = nested
        self.callback_args = tuple([nested] + list(self.callback_args))

    def deserialize(self, owner, val):
        return self.nested_field.deserialize(owner, val)

    def serialize(self, owner, obj):
        return self.nested_field.serialize(owner, obj)


class Alias(Callback):
    """A shallow alias to another field. """

    def __init__(self, field_name):
        """
        Alias field initialize. Exposes a shallow alias to another field that can be accessed by
        a different attribute key.

        :param field_name: the key of the other field
        :type field_name: basestring
        """

        self.alias = field_name
        super().__init__(
            self.alias_callback,
            callback_args=(Callback.SELF, self.alias),
            always_dump=True,
            data_key=field_name,
        )

    @staticmethod
    def alias_callback(m, field_name):
        return getattr(m, field_name)



import asyncio
from functools import wraps, partial
from tqdm import tqdm
import itertools


def chunkify(list, n):
    """Breaks a list into n chunks. Last chunk may not be equal in size to other chunks"""
    l = list
    return [l[i : i + n] for i in range(0, len(l), n)]


def with_index(fxn):
    @wraps(fxn)
    def wrapped(i, *args, **kwargs):
        return (i, fxn(*args, **kwargs))

    return wrapped


async def exec_async_fxn(
    fxn, arg_list, kwargs, chunk_size=1, desc="", progress_bar=True
):
    """
    Executes an asynchronous function

    :param fxn: function to execute in parallel
    :type fxn: callable
    :param arg_list: list of arguments as tuples for each function run
    :type arg_list: list of tuples
    :param kwargs: key-value arguments for each function run
    :type kwargs: dict
    :param chunk_size: number of functions to run for each worker
    :type chunk_size: int
    :param desc: description
    :type desc: basestring
    :param progress_bar: whether to display tqdm progress bar
    :type progress_bar: boolean
    :return: list of results in same order as arg_list
    :rtype: list
    """
    # run asynchronously

    loop = asyncio.get_event_loop()
    partial_fxn = partial(with_index(fxn), **kwargs)
    futures = [
        loop.run_in_executor(None, partial_fxn, i, *args)
        for i, args in enumerate(arg_list)
    ]

    # collect results
    results = []

    iterator = asyncio.as_completed(futures)
    if progress_bar:
        iterator = tqdm(iterator, desc=desc, total=len(futures), unit_scale=chunk_size)

    for f in iterator:
        results.append(await f)
    results = sorted(results, key=lambda x: x[0])
    return [r[1] for r in results]


def asyncfunc(fxn, arg_list, kwargs=None, chunk_size=1, progress_bar=True, desc=None):
    """
    Runs a function asynchronously.

    :param fxn: function to run asynchronously
    :type fxn: function or lambda
    :param arg_chunks: arguments to apply to the function; suggested to divide list into chunks
    :type arg_chunks: list
    :return: result
    :rtype: list
    """
    if kwargs is None:
        kwargs = {}
    # finish loop
    loop = asyncio.get_event_loop()
    if desc is None:
        desc = desc
    results = loop.run_until_complete(
        exec_async_fxn(
            fxn,
            arg_list,
            kwargs=kwargs,
            desc=desc,
            progress_bar=progress_bar,
            chunk_size=chunk_size,
        )
    )
    #     loop.close()
    return results


def make_async(chunk_size, progress_bar=True, as_classmethod=False, data_pos=0):
    """
    Wrapper to make a function run asynchrounously.

    :param chunk_size: size of array to apply to each worker
    :type chunk_size: int
    :param progress_bar: whether to display a progress bar
    :type progress_bar: bool
    :param as_classmethod: whether to pass in the first argument as a instance for instance or classmethods
    :type as_classmethod: bool
    :param data_pos: position in arguments where list of data is
    :type data_pos: int
    :return: results
    :rtype: list
    """

    def dec(fxn):
        data_position = data_pos
        if as_classmethod:
            data_position += 1

        @wraps(fxn)
        def wrapper(*args, **kwargs):
            data = args[data_position]
            post_data_args = args[data_position + 1 :]
            pre_data_args = args[:data_position]
            chunks = chunkify(data, chunk_size)
            arg_list = [pre_data_args + (c,) + post_data_args for c in chunks]
            desc = 'Running "{}" [size: {}, num: {}]: '.format(
                fxn.__name__, chunk_size, len(chunks)
            )
            results = asyncfunc(
                fxn, arg_list, kwargs=kwargs, progress_bar=progress_bar, desc=desc
            )
            return list(itertools.chain(*results))

        return wrapper

    return dec



"""Aquarium experiment planner

==========
planner
==========

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    planner
    layout

"""

from pydent.planner.planner import Planner, PlannerException
from pydent.planner.layout import PlannerLayout



"""Aquarium models

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    inventory
    sample
    field_value
    data_associations
    plan
    code
    operation
    job
    user
    crud_mixin
    field_value_mixins

This module contains a set of classes for various various Aquarium objects.
Trident models inherit the ModelBase class and have a model schema
(generated by ``@add_schema``) that handles the JSON loading and dumping.

By default, Trident models capture ALL JSON attribute/values and sets
the attributes of the resultant object.

.. code-block:: python

    u = User.load({"id": 1, "login": "John"})
    u.login    # => "John"
    u.id       # => 1

Fields and field options are added as class variables in the model class
definition.

Trident models also have specialized `relationships` which with other models,
which are initialized in the class definition as follows:

.. code-block:: python

    @add_schema
    class SampleType(Base):
        fields = {
            "samples": HasMany("Sample", "SampleType")
        }

When called as an attribute, Trident will automatically use the session
connection with Aquarium to retrieve the given model. This result is,
by default, cached in the model instance. For example, the following
code collects a single :class:`pydent.models.Sample` from the server, and then makes a new
requests for its :class:`pydent.models.SampleType`:

.. code-block:: python

    sample = session.Sample.one()
    sample_type = sample.sample_type         # new request made here for SampleType
    assert sample_type is sample.sample_type # no new request made here, cached result is returned

The cached result of relationships can be reset by setting the field attribute to `None`:

.. code-block:: python

    sample.sample_type = None  # reset cache for 'sample_type'
    st = sample.sample_type    # new request made here

More details on field definitions can be found at :class:`pydent.relationships`.

When called as an attribute, Trident will automatically use the session
connection with Aquarium to retrieve the given model.

For example, the following will define that SampleType has many Samples.
When .samples is called on a SampleType instance, Trident will use the database
to retrieve all samples that have a sample_type_id equal to the id of the
SampleType:

.. code-block:: python

    @add_schema
    class SampleType(Base):
        samples = Many("Sample",
            callback_args={"sample_type_id": lambda self: self.id})
"""

from pydent.models.code import Code, Library
from pydent.models.data_associations import DataAssociation, Upload
from pydent.models.field_value import FieldValue, FieldType, AllowableFieldType
from pydent.models.inventory import Item, Collection, ObjectType, PartAssociation
from pydent.models.job import Job, JobAssociation
from pydent.models.operation import Operation, OperationType
from pydent.models.plan import Plan, PlanAssociation, Wire
from pydent.models.sample import Sample, SampleType
from pydent.models.user import (
    Account,
    Budget,
    Group,
    Invoice,
    Membership,
    User,
    UserBudgetAssociation,
)

__all__ = [
    "Account",
    "AllowableFieldType",
    "Budget",
    "Code",
    "Collection",
    "DataAssociation",
    "FieldType",
    "FieldValue",
    "Group",
    "Invoice",
    "Item",
    "Job",
    "JobAssociation",
    "Library",
    "Membership",
    "ObjectType",
    "Operation",
    "OperationType",
    "PartAssociation",
    "Plan",
    "PlanAssociation",
    "Sample",
    "SampleType",
    "Upload",
    "User",
    "UserBudgetAssociation",
    "Wire",
]



from abc import ABC


class SessionABC(ABC):

    Account = None
    AllowableFieldType = None
    Budget = None
    Code = None
    Collection = None
    DataAssociation = None
    FieldType = None
    FieldValue = None
    Group = None
    Invoice = None
    Item = None
    Job = None
    JobAssociation = None
    Library = None
    Membership = None
    ObjectType = None
    Operation = None
    OperationType = None
    PartAssociation = None
    Plan = None
    PlanAssociation = None
    Sample = None
    SampleType = None
    Upload = None
    User = None
    UserBudgetAssociation = None
    Wire = None



"""
Models related to data associations
"""

import os
import shutil

import requests

from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.relationships import HasOne, JSON
from pydent.utils import make_async
from pydent.models.crud_mixin import JSONDeleteMixin, JSONSaveMixin
import json


class DataAssociatorMixin:
    """
    Mixin for handling data associations
    """

    def associate(self, key, value, upload=None):
        """
        Adds a data association with the key and value to this object.

        :param key: Key of the association
        :type key: str
        :param value: a json serializable object
        :type value: dict | str | int | float
        :param upload: optional file to upload
        :type upload: File
        :return: newly created data association
        :rtype: DataAssociation
        """
        """
        Adds a data association with the key and value to this object.
        """
        return self.session.utils.create_data_association(
            self, key, value, upload=upload
        )

    def associate_file(self, key, value, file, job_id=None):
        """
        Associate a file

        :param key: association key
        :type key: str or json
        :param value: association value
        :type value: str or json
        :param file: file to create :class:`Upload`
        :type file: file object
        :param job_id: optional job_id to associate the :class:`Upload`
        :type job_id: int
        :return: new data association
        :rtype: :class:`DataAssociation`
        """
        u = self.session.Upload.new(job_id=job_id, file=file)
        u.save()
        return self.associate(key, value, upload=u)

    def associate_file_from_path(self, key, value, filepath, job_id=None):
        """
        Associate a file from a filepath

        :param key: association key
        :type key: str or json
        :param value: association value
        :type value: str or json
        :param filepath: path to file to create :class:`Upload`
        :type filepath: str
        :param job_id: optional job_id to associate the :class:`Upload`
        :type job_id: int
        :return: new data association
        :rtype: :class:`DataAssociation`
        """
        with open(filepath, "rb") as f:
            return self.associate_file(key, value, f, job_id=job_id)

    def get_data_associations(self, key):
        das = []
        for da in self.data_associations:
            if da.key == key:
                das.append(da)
        return das

    # TODO: DataAssociation - do we really want to have this return either a list or single value? How are people using this?
    def get(self, key):
        val = []
        for da in self.get_data_associations(key):
            val.append(da.value)
        if len(val) == 1:
            return val[0]
        elif len(val) == 0:
            return None
        return val


@add_schema
class DataAssociation(JSONDeleteMixin, ModelBase):
    """A DataAssociation model"""

    fields = dict(object=JSON(), upload=HasOne("Upload"))

    @property
    def value(self):
        return self.object.get(self.key, None)

    def __str__(self):
        return self._to_str("id", "object")


@add_schema
class Upload(ModelBase):
    """
    An Upload model
    """

    fields = dict(job=HasOne("Job"))

    def __init__(self, job_id=None, file=None):
        """
        Create a new upload

        :param job_id: job id to associate the upload to
        :type job_id: int
        :param file: file to upload
        :type file: file object
        """
        super().__init__(job_id=job_id)
        self.file = file

    query_hook = dict(methods=["size", "name", "job"])

    # def _get_uploads_from_job_id(self, job_id):

    def _get_uploads_from_job(self):
        http = self.session._AqSession__aqhttp
        return http.get("krill/uploads?job={}".format(self.job_id))["uploads"]

    def temp_url(self):
        data = self.session.Upload.where({"id": self.id}, methods=["expiring_url"])[
            0
        ].raw
        return data["expiring_url"]

    @staticmethod
    def _download_file_from_url(url, outpath):
        """
        Downloads a file from a url

        :param url: url of file
        :type url: str
        :param outpath: filepath of out file
        :type outpath: str
        :return: http response
        :rtype: str
        """
        response = requests.get(url, stream=True)
        with open(outpath, "wb") as out_file:
            shutil.copyfileobj(response.raw, out_file)
        return response.raw

    @staticmethod
    @make_async(1)
    def async_download(uploads, outdir=None, overwrite=True):
        """
        Asynchronously downloads from list of :class:`Upload` models.

        :param uploads: list of Uploads
        :type uploads: list
        :param outdir: path to output directory to save downloaded files
        :type outdir: str
        :param overwrite: if True, will overwrite existing files
        :type overwrite: bool
        :return: list of filepaths
        :rtype: list
        """
        return Upload._download_files(uploads, outdir, overwrite)

    @staticmethod
    def _download_files(uploads, outdir, overwrite):
        """
        Downloads uploaded file from list of :class:`Upload` models.

        :param uploads: list of Uploads
        :type uploads: list
        :param outdir: path to output directory to save downloaded files (defaults to current directory)
        :type outdir: str
        :param overwrite: if True, will overwrite existing files (default: True)
        :type overwrite: bool
        :return: list of filepaths
        :rtype: list
        """
        filepaths = []
        for upload in uploads:
            filepath = upload.download(outdir=outdir, overwrite=overwrite)
            filepaths.append(filepath)
        return filepaths

    def download(self, outdir=None, filename=None, overwrite=True):
        """
        Downloads the uploaded file to the specified output directory. If
        no output directory is specified, the file will be downloaded to the
        current directory.

        :param outdir: path of directory of output file (default is current directory)
        :param outfile: filename of output file (defaults to upload_filename)
        :param overwrite: whether to overwrite file if it already exists
        :return: None
        """
        if outdir is None:
            outdir = "."
        if filename is None:
            filename = "{}_{}".format(self.id, self.upload_file_name)
        filepath = os.path.join(outdir, filename)
        if not os.path.exists(filepath) or overwrite:
            self._download_file_from_url(self.temp_url(), filepath)
        return filepath

    @property
    def data(self):
        """Return the data associated with the upload."""
        result = requests.get(self.temp_url())
        return result.content

    def create(self):
        """Save the upload to the server."""
        return self.session.utils.create_upload(self)

    def save(self):
        return self.create()
